from ..logger import logger
logger.debug("Started 'reading app/settings_editor.py'")

import numpy as np
from qtpy import QtWidgets, QtGui
from multiprocessing import cpu_count
from ..utils.misc import setConsoleVisible
from ..utils.BaseProcess import BaseDialog, BaseProcess, ColorSelector
from .. import global_vars as g


__all__ = ['SettingsEditor', 'rectSettings', 'pointSettings', 'pencilSettings']

data_types = ['uint8', 'uint16', 'uint32', 'uint64', 'int8', 'int16', 'int32', 'int64', 'float16', 'float32', 'float64']

class SettingsEditor(BaseDialog):
    gui = None
    def __init__(self):
        from pyqtgraph import ComboBox

        old_dtype=g.settings['internal_data_type']
        dataDrop = ComboBox(items=data_types, default=old_dtype)
        showCheck = QtWidgets.QCheckBox()
        showCheck.setChecked(g.settings['show_windows'])
        multipleTracesCheck = QtWidgets.QCheckBox()
        multipleTracesCheck.setChecked(g.settings['multipleTraceWindows'])
        multiprocessing = QtWidgets.QCheckBox()
        multiprocessing.setChecked(g.settings['multiprocessing'])
        nCores = QtWidgets.QComboBox()
        debug_check = QtWidgets.QCheckBox(checked=g.settings['debug_mode'])
        debug_check.toggled.connect(setConsoleVisible)
        for i in np.arange(cpu_count())+1:
            nCores.addItem(str(i))
        nCores.setCurrentIndex(g.settings['nCores']-1)
        random_color_check = QtWidgets.QCheckBox()
        random_color_check.setChecked(g.settings['roi_color'] == 'random')
        roi_color = ColorSelector()
        roi_color.setEnabled(g.settings['roi_color'] != 'random')
        roi_color.color=g.settings['roi_color'] if g.settings['roi_color'] != 'random' else '#ffff00'

        default_roi = QtWidgets.QCheckBox()
        default_roi.setChecked(g.settings['default_roi_on_click'])
        
        items = []
        items.append({'name': 'random_color_check', 'string': 'Randomly color ROIs', 'object': random_color_check})
        items.append({'name': 'roi_color', 'string': 'Default ROI Color', 'object': roi_color})
        items.append({'name': 'internal_data_type', 'string': 'Internal Data Type', 'object': dataDrop})
        items.append({'name': 'show_windows', 'string': 'Show Windows', 'object': showCheck})
        items.append({'name': 'multipleTraceWindows', 'string': 'Multiple Trace Windows', 'object': multipleTracesCheck})
        items.append({'name': 'multiprocessing', 'string': 'Multiprocessing On', 'object': multiprocessing})
        items.append({'name': 'nCores', 'string': 'Number of cores to use when multiprocessing', 'object': nCores})
        items.append({'name': 'debug_mode', 'string': 'Debug Mode', 'object': debug_check})
        items.append({'name': 'default_roi_on_click', 'string': 'Enable default ROI on right click', 'object': default_roi})
        
        def update():
            g.settings['internal_data_type'] = str(dataDrop.currentText())
            g.settings['show_windows'] = showCheck.isChecked()
            g.settings['multipleTraceWindows'] = multipleTracesCheck.isChecked()
            g.settings['multiprocessing']=multiprocessing.isChecked()
            g.settings['nCores']=int(nCores.itemText(nCores.currentIndex()))
            g.settings['debug_mode'] = debug_check.isChecked()
            if not random_color_check.isChecked() and roi_color.color == 'random':
                roi_color.color = "#ffff00"
            g.settings['roi_color'] = roi_color.value() if not random_color_check.isChecked() else "random"
            roi_color.setEnabled(g.settings['roi_color'] != 'random')
            g.settings['default_roi_on_click'] = default_roi.isChecked()
            

        super(SettingsEditor, self).__init__(items, 'flika settings', None)
        self.accepted.connect(update)
        self.changeSignal.connect(update)
        g.dialogs.append(self)

    @staticmethod
    def show():
        if SettingsEditor.gui == None:
            SettingsEditor.gui = SettingsEditor()
        BaseDialog.show(SettingsEditor.gui)


def pencilSettings(pencilButton):
    """
    default points color
    default points size
    currentWindow points color
    currentWindow points size
    clear current points

    """
    pencil_value = QtWidgets.QSpinBox()
    pencil_value.setRange(0, 1000)
    v = g.settings['pencil_value']
    if v is None:
        v = 0
    pencil_value.setValue(v)
    items = []
    items.append({'name': 'pencil_value', 'string': 'Pencil Value', 'object': pencil_value})
    def update_final():
        g.settings['pencil_value'] = pencil_value.value()
    dialog = BaseDialog(items, 'Points Settings', '')
    g.dialogs.append(dialog)
    dialog.accepted.connect(update_final)
    dialog.show()

def pointSettings(pointButton):
    """
    default points color
    default points size
    currentWindow points color
    currentWindow points size
    clear current points

    """
    point_color = ColorSelector()
    point_color.color=g.settings['point_color']
    point_color.label.setText(point_color.color)
    point_size = QtWidgets.QSpinBox()
    point_size.setRange(1, 50)
    point_size.setValue(g.settings['point_size'])
    show_all_points = QtWidgets.QCheckBox()
    show_all_points.setChecked(g.settings['show_all_points'])
    delete_all_points = QtWidgets.QCheckBox()
    delete_all_points.setChecked(False)
    
    update_current_points_check = QtWidgets.QCheckBox()
    update_current_points_check.setChecked(False)
    
    items = []
    items.append({'name': 'point_color', 'string': 'Default Point Color', 'object': point_color})
    items.append({'name': 'point_size', 'string': 'Default Point Size', 'object': point_size})
    items.append({'name': 'show_all_points', 'string': 'Show points from all frames', 'object': show_all_points})
    items.append({'name': 'update_current_points_check', 'string': 'Update already plotted points', 'object': update_current_points_check})
    items.append({'name': 'delete_all_points', 'string': 'Delete all points', 'object': delete_all_points})

    def update():
        from pyqtgraph import mkBrush
        win = g.win
        g.settings['point_color'] = point_color.value()
        g.settings['point_size'] = point_size.value()
        g.settings['show_all_points'] = show_all_points.isChecked()
        if win is not None and update_current_points_check.isChecked() == True:
            color = QtGui.QColor(point_color.value())
            size = point_size.value()
            for t in np.arange(win.mt):
                for i in np.arange(len(win.scatterPoints[t])):
                    win.scatterPoints[t][i][2] = color
                    win.scatterPoints[t][i][3] = size
            win.updateindex()
        if win is not None:
            if g.settings['show_all_points']:
                pts = []
                for t in np.arange(win.mt):
                    pts.extend(win.scatterPoints[t])
                pointSizes = [pt[3] for pt in pts]
                brushes = [mkBrush(*pt[2].getRgb()) for pt in pts]
                win.scatterPlot.setPoints(pos=pts, size=pointSizes, brush=brushes)
            else:
                win.updateindex()

    def update_final():
        win = g.win
        if win is not None and delete_all_points.isChecked():
            for t in np.arange(win.mt):
                win.scatterPoints[t] = []
            win.updateindex()
        update()

    dialog = BaseDialog(items, 'Points Settings', '')
    g.dialogs.append(dialog) 
    dialog.accepted.connect(update_final)
    dialog.changeSignal.connect(update)
    dialog.show()

def rectSettings(rectButton):
    """
    default rect height
    default rect width
    """
    rect_width = QtWidgets.QSpinBox()
    rect_width.setRange(1,500)
    rect_width.setValue(g.settings['rect_width'])
    rect_height = QtWidgets.QSpinBox()
    rect_height.setRange(1,500)
    rect_height.setValue(g.settings['rect_height'])

    items = []
    items.append({'name': 'rect_width', 'string': 'Default Rectangle Width', 'object': rect_width})
    items.append({'name': 'rect_height', 'string': 'Default Rectangle Height', 'object': rect_height})
    
    def update():
        win = g.win
        g.settings['rect_width'] = rect_width.value()
        g.settings['rect_height'] = rect_height.value()
        
    dialog = BaseDialog(items, 'Rectangle Settings', '')
    dialog.accepted.connect(update)
    dialog.changeSignal.connect(update)
    g.dialogs.append(dialog)
    dialog.show()

logger.debug("Completed 'reading app/settings_editor.py'")



# -*- coding: utf-8 -*-
from .logger import logger
logger.debug("Started 'reading tracefig.py'")
import os
import time
import numpy as np
from qtpy import QtCore, QtGui, QtWidgets
from scipy.fftpack import fft, fftfreq, fftshift
import pyqtgraph as pg
from pyqtgraph.dockarea import *
from . import global_vars as g
from .utils.misc import save_file_gui


pg.setConfigOptions(useWeave=False)


class TraceFig(QtWidgets.QWidget):
    """Pyqtgraph PlotWidget with frame range selector. Display average trace of ROIs and updates in realtime.
    
    Attributes:
        p1 (pg.PlotWidget): top plot widget displaying selected region of traces
        p2 (pg.PlotWidget): bottom plot widget displaying entire traces with region selection
        rois ([dict]): list of roi information dicts for reference

    Signals:
        :indexChanged(int): emits the index of the mouse when the user hovers over the top plotWidget
        :finishedDrawingSignal(): emits when the bottom ROI is finished updating
        :keyPressSignal(QtCore.QEvent): emits when the traceWindow is selected and a key is pressed
        :partialThreadUpdatedSignal(): emits when the top plot widget is updated
    """
    indexChanged=QtCore.Signal(int)
    finishedDrawingSignal=QtCore.Signal()
    keyPressSignal=QtCore.Signal(QtCore.QEvent)
    partialThreadUpdatedSignal = QtCore.Signal()
    name = "Trace Widget"

    def __init__(self):
        super(TraceFig, self).__init__()
        g.traceWindows.append(self)
        self.setCurrentTraceWindow()
        if 'tracefig_settings' in g.settings and 'coords' in g.settings['tracefig_settings']:
            self.setGeometry(QtCore.QRect(*g.settings['tracefig_settings']['coords']))
        else:
            self.setGeometry(QtCore.QRect(355, 30, 1219, 148))
        self.setWindowTitle('flika')
        self.l = QtWidgets.QVBoxLayout()

        self.l.setContentsMargins(0,0,0,0)
        self.setLayout(self.l)
        self.p1=pg.PlotWidget()
        self.p2=pg.PlotWidget()
        self.p1.getPlotItem().axes['left']['item'].setGrid(100) #this makes faint horizontal lines
        self.p2.setMaximumHeight(50)
        self.export_button = QtWidgets.QPushButton("Export")
        self.export_button.setMaximumWidth(100)
        self.export_button.clicked.connect(self.export_gui)
        self.power_spectrum_button = QtWidgets.QPushButton("Power Spectrum")
        self.power_spectrum_button.setMaximumWidth(100)
        self.power_spectrum_button.clicked.connect(self.generate_power_spectrum)
        self.button_layout = QtWidgets.QGridLayout()
        self.l.addWidget(self.p1, 1)
        self.l.addWidget(self.p2, 1)
        self.l.addLayout(self.button_layout)
        self.button_layout.setContentsMargins(0, 0, 0, 0)
        self.button_layout.addWidget(self.export_button, 0, 0)
        self.button_layout.addWidget(self.power_spectrum_button, 0, 1)
        verticalSpacer = QtWidgets.QSpacerItem(10, 10, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.button_layout.addItem(verticalSpacer, 0, 2)
    
        self.region = pg.LinearRegionItem()         # Add the LinearRegionItem to the ViewBox, but tell the ViewBox to exclude this item when doing auto-range calculations.
        self.region.setZValue(10)
        self.p2.plotItem.addItem(self.region, ignoreBounds=True)
        self.p1.setAutoVisible(y=True)
        self.rois=[] # roi in this list is a dict: {roi, p1trace,p2trace, sigproxy}
        self.redrawPartialThread = None
        self.vb = self.p1.plotItem.getViewBox()
        
        self.proxy = pg.SignalProxy(self.p1.scene().sigMouseMoved, rateLimit=60, slot=self.mouseMoved)
        self.p2.plotItem.vb.mouseDragEvent=self.mouseDragEvent2
        
        self.region.sigRegionChanged.connect(self.update)
        self.p1.plotItem.sigRangeChanged.connect(self.updateRegion)
        self.region.setRegion([0, 200])

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.p1.update)
        self.timer.start()

        from .process.measure import measure
        self.measure=measure
        self.p1.scene().sigMouseClicked.connect(self.measure.pointclicked)
        self.p1.scene().sigMouseClicked.connect(self.setCurrentTraceWindow)
        self.resizeEvent = self.onResize
        self.moveEvent = self.onMove
        
        if 'tracefig_settings' not in g.settings:
            g.settings['tracefig_settings']=dict()
            try:
                g.settings['tracefig_settings']['coords']=self.geometry().getRect()
            except Exception as e:
                g.alert(e)
        self.show()
        
    def onResize(self,event):
        g.settings['tracefig_settings']['coords']=self.geometry().getRect()

    def onMove(self,event):
        g.settings['tracefig_settings']['coords']=self.geometry().getRect()
        
    def setCurrentTraceWindow(self):
        if g.currentTrace is not None:
            g.currentTrace.setStyleSheet("border:1px solid rgb(0, 0, 0); ")
        self.setStyleSheet("border:1px solid rgb(0, 255, 0); ")
        g.currentTrace = self

    def mouseDragEvent2(self,ev):
        ev.ignore() # prevent anything from happening

    def mouseDragEvent1(self,ev):
        ev.ignore() # prevent anything from happening

    def keyPressEvent(self,ev):
        self.keyPressSignal.emit(ev)

    def closeEvent(self, event):
        while len(self.rois) > 0:
            self.removeROI(0)
        try:
            self.p1.scene().sigMouseClicked.disconnect(self.measure.pointclicked)
            self.p1.scene().sigMouseClicked.disconnect(self.setCurrentTraceWindow)
        except:
            pass
        if self in g.traceWindows:
            g.traceWindows.remove(self)
        g.currentTrace = None
        event.accept() # let the window close

    def update(self):
        self.region.setZValue(10)
        minX, maxX = self.region.getRegion()
        self.p1.plotItem.setXRange(minX, maxX, padding=0, update=False)    
        self.p1.plotItem.axes['bottom']['item'].setRange(minX,maxX)

    def updateRegion(self,window, viewRange):
        rgn = viewRange[0]
        self.region.setRegion(rgn)

    def getBounds(self):
        bounds=self.region.getRegion()
        bounds=[int(np.floor(bounds[0])),int(np.ceil(bounds[1]))+1]
        return bounds

    def mouseMoved(self,evt):
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ShiftModifier:
            pass
        else:
            pos = evt[0]  ## using signal proxy turns original arguments into a tuple
            if self.p1.plotItem.sceneBoundingRect().contains(pos):
                mousePoint = self.vb.mapSceneToView(pos)
                index = int(mousePoint.x())
                if index >= 0:
                    #self.label.setText("<span style='font-size: 12pt'>frame={0}</span>".format(index))
                    self.indexChanged.emit(index)
                    g.m.statusBar().showMessage('frame {}    y={}'.format(index,mousePoint.y()))
    
    def get_roi_index(self,roi):
        return [r['roi'] for r in self.rois].index(roi)
        
    def alert(self, msg):
        #print(msg)
        pass

    def translated(self,roi):
        index=self.get_roi_index(roi)
        self.rois[index]['toBeRedrawn']=True
        if self.redrawPartialThread is None or self.redrawPartialThread.isFinished():
            self.alert('Launching redrawPartialThread')
            self.redrawPartialThread = RedrawPartialThread(self)
            self.redrawPartialThread.alert.connect(self.alert)
            self.redrawPartialThread.start()
            self.redrawPartialThread.updated.connect(self.partialThreadUpdatedSignal.emit)
            
    def translateFinished(self, roi):
        roi_index = self.get_roi_index(roi)
        if self.redrawPartialThread is not None and self.redrawPartialThread.isRunning():
            self.redrawPartialThread.quit_loop = True
            #self.redrawPartialThread.finished_sig.emit() #tell the thread to finish
            #loop = QtCore.QEventLoop()
            #self.redrawPartialThread.finished.connect(loop.quit)
            #loop.exec_()# This blocks until the "finished" signal is emitted
        trace=roi.getTrace()
        self.update_trace_full(roi_index, trace)

    def update_trace_full(self, roi_index, trace):
        pen=QtGui.QPen(self.rois[roi_index]['roi'].pen)
        self.rois[roi_index]['p1trace'].setData(trace,pen=pen)
        self.rois[roi_index]['p2trace'].setData(trace,pen=pen)
        self.finishedDrawingSignal.emit()
        
    def addROI(self,roi):
        if self.hasROI(roi):
            return
        trace=roi.getTrace()
        if trace is None:
            raise InvalidTraceException()
        pen=QtGui.QPen(roi.pen)
        pen.setWidth(0)
        if len(trace)==1:
            p1trace=self.p1.plot(trace, pen=None, symbol='o')
            p2trace=self.p2.plot(trace, pen=None, symbol='o')
        else:
            p1trace=self.p1.plot(trace, pen=pen)
            p2trace=self.p2.plot(trace, pen=pen) 
        
        roi.sigRegionChanged.connect(self.translated)
        roi.sigRegionChangeFinished.connect(self.translateFinished)

        if len(self.rois) == 0:
            self.region.setRegion([0, len(trace)-1])
        self.rois.append(dict({'roi':roi,'p1trace':p1trace,'p2trace':p2trace,'toBeRedrawn':False,'toBeRedrawnFull':False}))

    def removeROI(self,roi):
        from .roi import ROI_Base
        if isinstance(roi, ROI_Base):
            index=[r['roi'] for r in self.rois].index(roi) #this is the index of the roi in self.rois
        elif isinstance(roi, int):
            index = roi
        else:
            g.alert("Failed to remove roi {}".format(roi))
            return
        self.p1.removeItem(self.rois[index]['p1trace'])
        self.p2.removeItem(self.rois[index]['p2trace'])
        self.rois[index]['roi'].traceWindow = None
        try:
            self.rois[index]['roi'].resetSignals()
        except:
            pass
        del self.rois[index]
        if len(self.rois)==0:
            self.close()

    def hasROI(self,roi):
        return roi in [r['roi'] for r in self.rois] #return True if roi is plotted

    def export_gui(self):
        filename = g.settings['filename']
        directory = os.path.dirname(filename)
        if filename is not None:
            filename = save_file_gui('Save Traces', directory, '*.txt')
        else:
            filename = save_file_gui('Save Traces', '', '*.txt')
        if filename == '':
            return False
        else:
            self.export(filename)

    def export(self,filename):
        ''' This function saves out all the traces in the tracefig to a file specified by the argument 'filename'.
        The output file is a tab seperated ascii file where each column is a trace.  
        Traces are saved in the order they were added to the plot.
        
        '''
        g.m.statusBar().showMessage('Saving {}'.format(os.path.basename(filename)))
        traces=[]
        for roi in self.rois:
            traces.append(roi['roi'].getTrace())
        traces.insert(0,np.arange(len(traces[0])))
        traces=np.array(traces).T
        np.savetxt(filename,traces,delimiter='\t',fmt='%10f')
        g.m.statusBar().showMessage('Successfully saved {}'.format(os.path.basename(filename)))

    def generate_power_spectrum(self):
        sample_interval = 1
        self.fft_analyzer = FFT_Analyzer(self.rois, sample_interval, self)

class FFT_Analyzer(QtWidgets.QWidget):
    def __init__(self, rois, sample_interval, tracefig, parent = None) :
        QtWidgets.QWidget.__init__(self, parent)
        """
        sample_interval is the sample_duration in seconds. If the sample is 1000 Hz, sample_interval = .001 (1 ms)
        """
        self.tracefig = tracefig
        self.rois = rois
        geo = self.tracefig.geometry()
        geo.adjust(0, geo.height(), 0, geo.height())
        self.setGeometry(geo)
        self.setWindowTitle("Power Spectrum")
        self.l = QtWidgets.QVBoxLayout()
        self.setLayout(self.l)
        self.area = DockArea()
        self.l.addWidget(self.area)

        self.d3 = Dock("FFT")
        self.area.addDock(self.d3, size=(382, 216))
        self.fftplt = pg.PlotWidget()
        self.d3.addWidget(self.fftplt)
        self.fftplt.showGrid(x=True, y=True)
        self.fftplt.setLogMode(x=True,y=True)
        self.fftplt.setLabel('bottom', 'Frequency')
        self.fftplt.setLabel('left', 'Power')
        self.sample_interval = sample_interval
        self.set_data(rois, self.sample_interval)

        self.export_button = QtWidgets.QPushButton("Export")
        self.export_button.setMaximumWidth(100)
        self.export_button.clicked.connect(self.export_gui)
        self.d3.addWidget(self.export_button)
        self.show()

    def set_data(self, rois, sample_interval):
        traces = []
        pens = []
        for roi in rois:
            traces.append(roi['roi'].getTrace())
            pen = QtGui.QPen(roi['roi'].pen)
            pen.setWidth(0)
            pens.append(pen)

        longest_trace_len = np.max([len(trace) for trace in traces])
        N = int(2**np.floor(np.log2(longest_trace_len)))
        # x = np.linspace(0.0, N * sample_interval, N)
        for i in np.arange(len(rois)):
            trace = traces[i]
            yf = fft(trace[-N:])
            xf = fftfreq(N, sample_interval)
            xf = xf[1:int(N / 2)]
            yf = np.abs(yf[1:int(N / 2)])**2
            rois[i]['power_spectrum_x'] = xf
            rois[i]['power_spectrum_y'] = yf
            self.fftplt.plot(xf, yf, pen=pens[i])

    def export_gui(self):
        filename = g.settings['filename']
        directory = os.path.dirname(filename)
        if filename is not None:
            filename = save_file_gui('Save Power Spectrum', directory, '*.csv')
        else:
            filename = save_file_gui('Save Power Spectrum', '', '*.csv')
        if filename == '':
            return False
        else:
            self.export(filename)

    def export(self, filename):
        ''' This function saves out all the traces in the 'Power Spectrum" to a file specified by the argument 'filename'.
        The output file is a csv.
        Traces are saved in the order they were added to the tracefig window.

        '''
        g.m.statusBar().showMessage('Saving {}'.format(os.path.basename(filename)))
        traces = []
        headers = []
        cols = []
        for i, roi in enumerate(self.rois):
            cols.append(roi['power_spectrum_x'])
            cols.append(roi['power_spectrum_y'])
            headers.append('X roi{}'.format(i))
            headers.append('Y roi{}'.format(i))
        header = ','.join(headers)
        cols = np.array(cols).T
        np.savetxt(filename, cols, header=header, delimiter=',', comments='', fmt='%10f')
        g.m.statusBar().showMessage('Successfully saved {}'.format(os.path.basename(filename)))


def roiPlot(roi):
    '''
    returns tracefig that is used to plot roi
    '''
    if g.settings['multipleTraceWindows'] or g.currentTrace is None:
        win = TraceFig()
    else:
        win = g.currentTrace
    try:
        win.addROI(roi)
    except InvalidTraceException:
        if len(win.rois) == 0:
            win.close()
            return None
    return win


class RedrawPartialThread(QtCore.QThread):
    finished=QtCore.Signal() #this announces that the thread has finished
    finished_sig=QtCore.Signal() #This tells the thread to finish
    alert = QtCore.Signal(str)
    updated = QtCore.Signal() #This signal is emitted after each redraw

    def __init__(self,tracefig):
        QtCore.QThread.__init__(self)
        self.tracefig=tracefig
        self.redrawCompleted=True
        self.quit_loop=False
        
    def run(self):
        self.finished_sig.connect(self.request_quit_loop)
        while self.quit_loop is False:
            time.sleep(.05)
            self.redraw()
            self.updated.emit()
        self.alert.emit("Finished Redraw")
        self.finished.emit()
        
    def request_quit_loop(self):
        self.quit_loop=True
        
    def redraw(self):
        if self.redrawCompleted is False:
            self.alert.emit("Redraw hasn't finished")
            pass
        else:
            self.alert.emit("Redrawing")
            self.redrawCompleted=False
            idxs=[]
            for i in np.arange(len(self.tracefig.rois)):
                if self.tracefig.rois[i]['toBeRedrawn']:
                    self.tracefig.rois[i]['toBeRedrawn']=False
                    idxs.append(i)
            traces=[]
            bounds=self.tracefig.getBounds()
            bounds = [max(0, bounds[0]), bounds[1]]
            for i in idxs:
                roi=self.tracefig.rois[i]['roi']
                trace=roi.getTrace(bounds)
                traces.append(trace)
            for i, roi_index in enumerate(idxs):
                trace=traces[i] #This function can sometimes take a long time.  
                pen=QtGui.QPen(self.tracefig.rois[roi_index]['roi'].pen)
                bb=self.tracefig.getBounds()
                curve=self.tracefig.rois[roi_index]['p1trace']
                newtrace=curve.getData()[1]
                if bb[0]<0: bb[0]=0
                if bb[1]>len(newtrace): bb[1]=len(newtrace)
                if bb[1]<0 or bb[0]>len(newtrace):
                    return
                newtrace[bb[0]:bb[1]]=trace
                curve.setData(newtrace,pen=pen)
                self.alert.emit("CURVE {} redrawn".format(roi_index))
                QtWidgets.qApp.processEvents()
            self.redrawCompleted=True

class InvalidTraceException(Exception):
    pass


logger.debug("Completed 'reading tracefig.py'")



# -*- coding: utf-8 -*-
from .logger import logger
logger.debug("Started 'reading window.py'")
from qtpy import QtCore, QtGui, QtWidgets
import pyqtgraph as pg
import os, time
import numpy as np
from . import global_vars as g
from .roi import *
from .utils.misc import save_file_gui
from .utils.BaseProcess import WindowSelector, SliderLabel

pg.setConfigOptions(useWeave=False)


class Bg_im_dialog(QtWidgets.QDialog):
    def __init__(self, parent=None):
        QtWidgets.QDialog.__init__(self)
        self.parent = parent
        self.setWindowTitle("Select background image")
        self.window_selector = WindowSelector()
        self.window_selector.valueChanged.connect(self.bg_win_changed)
        self.alpha_slider = SliderLabel(3)
        self.alpha_slider.setRange(0,1)
        self.alpha_slider.setValue(.5)
        self.alpha_slider.valueChanged.connect(self.alpha_changed)
        self.formlayout = QtWidgets.QFormLayout()
        self.formlayout.setLabelAlignment(QtCore.Qt.AlignRight)
        self.formlayout.addRow("Select window with background image", self.window_selector)
        self.formlayout.addRow("Set background opacity", self.alpha_slider)
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addLayout(self.formlayout)
        self.setLayout(self.layout)

    def alpha_changed(self, value):
        self.parent.bg_im.setOpacity(value)

    def bg_win_changed(self):
        if self.parent.bg_im is not None:
            self.parent.imageview.view.removeItem(self.parent.bg_im)
            self.bg_im = None
        self.parent.bg_im = pg.ImageItem(self.window_selector.window.imageview.imageItem.image)
        self.parent.bg_im.setOpacity(self.alpha_slider.value())
        self.parent.imageview.view.addItem(self.parent.bg_im)


    def closeEvent(self,ev):
        if self.parent.bg_im is not None:
            self.parent.imageview.view.removeItem(self.parent.bg_im)
            self.bg_im = None



class ImageView(pg.ImageView):
    def __init__(self, *args, **kargs):
        pg.ImageView.__init__(self, *args, **kargs)
        self.view.unregister()
        self.view.removeItem(self.roi)
        self.view.removeItem(self.normRoi)
        self.roi.setParent(None)
        self.normRoi.setParent(None)
        self.ui.menuBtn.setParent(None)
        self.ui.roiBtn.setParent(None) # gets rid of 'roi' button that comes with ImageView
        self.ui.normLUTbtn = QtWidgets.QPushButton(self.ui.layoutWidget)
        self.ui.normLUTbtn.setObjectName("LUT norm")
        self.ui.normLUTbtn.setText("LUT norm")
        self.ui.gridLayout.addWidget(self.ui.normLUTbtn, 1, 1, 1, 1)

        self.ui.bg_imbtn = QtWidgets.QPushButton(self.ui.layoutWidget)
        self.ui.bg_imbtn.setObjectName("bg im")
        self.ui.bg_imbtn.setText("bg im")
        self.ui.gridLayout.addWidget(self.ui.bg_imbtn, 1, 2, 1, 1)

        self.ui.roiPlot.setMaximumHeight(40)
        self.ui.roiPlot.getPlotItem().getViewBox().setMouseEnabled(False)
        self.ui.roiPlot.getPlotItem().hideButtons()

    def hasTimeAxis(self):
        return 't' in self.axes and not (self.axes['t'] is None or self.image.shape[self.axes['t']] == 1)

    def roiClicked(self):
        showRoiPlot = False
        if self.hasTimeAxis():
            showRoiPlot = True
            mn = self.tVals.min()
            mx = self.tVals.max() + .01
            self.ui.roiPlot.setXRange(mn, mx, padding=0.01)
            self.timeLine.show()
            self.timeLine.setBounds([mn, mx])
            self.ui.roiPlot.show()
            if not self.ui.roiBtn.isChecked():
                self.ui.splitter.setSizes([self.height()-35, 35])
        else:
            self.timeLine.hide()
            #self.ui.roiPlot.hide()
            
        self.ui.roiPlot.setVisible(showRoiPlot)

class Window(QtWidgets.QWidget):
    """
    Window objects are the central objects in flika. Almost all functions in the 
    :mod:`process <flika.process>` module are performed on Window objects and 
    output Window objects. 

    Args:
        tif (numpy.array): The image the window will store and display
        name (str): The name of the window.
        filename (str): The filename (including full path) of file this window's image orinated from.
        commands (list of str): a list of the commands used to create this window, starting with loading the file.
        metadata (dict): dict: a dictionary containing the original file's metadata.


    """
    closeSignal = QtCore.Signal()
    keyPressSignal = QtCore.Signal(QtCore.QEvent)
    sigTimeChanged = QtCore.Signal(int)
    gainedFocusSignal = QtCore.Signal()
    lostFocusSignal = QtCore.Signal()

    def __init__(self, tif, name='flika', filename='', commands=[], metadata=dict()):
        from .process.measure import measure
        QtWidgets.QWidget.__init__(self)
        self.name = name  #: str: The name of the window.
        self.filename = filename  #: str: The filename (including full path) of file this window's image orinated from.
        self.commands = commands  #: list of str: a list of the commands used to create this window, starting with loading the file.
        self.metadata = metadata  #: dict: a dictionary containing the original file's metadata.
        self.volume = None  # When attaching a 4D array to this Window object, where self.image is a 3D slice of this volume, attach it here. This will remain None for all 3D Windows
        self.scatterPlot = None
        self.closed = False  #: bool: True if the window has been closed, False otherwise.
        self.mx = 0  #: int: The number of pixels wide the image is in the x (left to right) dimension.
        self.my = 0  #: int: The number of pixels heigh the image is in the y (up to down) dimension.
        self.mt = 0  #: int: The number of frames in the image stack.
        self.framerate = None  #: float: The number of frames per second (Hz).
        self.image = tif
        self.dtype = tif.dtype  #: dtype: The datatype of the stored image, e.g. ``uint8``.
        self.top_left_label = None
        self.rois = []  #: list of ROIs: a list of all the :class:`ROIs <flika.roi.ROI_Base>` inside this window.
        self.currentROI = None  #: :class:`ROI <flika.roi.ROI_Base>`: When an ROI is clicked, it becomes the currentROI of that window and can be accessed via this variable.
        self.creatingROI = False
        self.imageview = None
        self.bg_im = None
        self.currentIndex = 0
        self.linkedWindows = set()
        self.measure = measure
        self.resizeEvent = self.onResize
        self.moveEvent = self.onMove
        self.pasteAct = QtWidgets.QAction("&Paste", self, triggered=self.paste)
        self.sigTimeChanged.connect(self.showFrame)
        self._check_for_infinities(tif)
        self._init_dimensions(tif)
        self._init_imageview(tif)
        self.setWindowTitle(name)
        self.normLUT(tif)
        self._init_scatterplot()
        self._init_menu()
        self._init_geometry()
        self.setAsCurrentWindow()

    def _init_geometry(self):
        assert g.win != self  # self.setAsCurrentWindow() must be called after this function
        if g.win is None:
            if 'window_settings' not in g.settings:
                g.settings['window_settings'] = dict()
            if 'coords' in g.settings['window_settings']:
                geometry = QtCore.QRect(*g.settings['window_settings']['coords'])
            else:
                width = 684
                height = 585
                nwindows = len(g.windows)
                x = 10 + 10 * nwindows
                y = 300 + 10 * nwindows
                geometry = QtCore.QRect(x, y, width, height)
                g.settings['window_settings']['coords'] = geometry.getRect()
        else:
            geometry = g.win.geometry()

        desktopGeom = QtWidgets.QDesktopWidget().screenGeometry()
        maxX = (desktopGeom.width() - geometry.width()) or 1
        maxY = (desktopGeom.height() - geometry.height()) or 1
        newX = (geometry.x() + 10) % maxX
        newY = ((geometry.y() + 10) % maxY) or 30
        
        geometry = QtCore.QRect(newX, newY, geometry.width(), geometry.height())
        self.setGeometry(geometry)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.addWidget(self.imageview)
        self.layout.setContentsMargins(0, 0, 0, 0)
        if g.settings['show_windows']:
            self.show()
            self.raise_()
            QtWidgets.qApp.processEvents()

    def _check_for_infinities(self, tif):
        try:
            if np.any(np.isinf(tif)):
                tif[np.isinf(tif)] = 0
                g.alert('Some array values were inf. Setting those values to 0')
        except MemoryError:
            pass

    def _init_imageview(self, tif):
        self.imageview = ImageView(self)
        self.imageview.setMouseTracking(True)
        self.imageview.installEventFilter(self)
        self.imageview.ui.normLUTbtn.pressed.connect(lambda: self.normLUT(self.image))
        self.imageview.ui.bg_imbtn.pressed.connect(self.set_bg_im)
        rp = self.imageview.ui.roiPlot.getPlotItem()
        self.linkMenu = QtWidgets.QMenu("Link frame")
        rp.ctrlMenu = self.linkMenu
        self.linkMenu.aboutToShow.connect(self.make_link_menu)
        self.imageview.setImage(tif)
        def clicked(evt):
            self.measure.pointclicked(evt, window=self)
        self.imageview.scene.sigMouseClicked.connect(clicked)
        self.imageview.timeLine.sigPositionChanged.connect(self.updateindex)
        self.currentIndex = self.imageview.currentIndex
        self.imageview.scene.sigMouseMoved.connect(self.mouseMoved)
        self.imageview.view.mouseDragEvent = self.mouseDragEvent
        self.imageview.view.mouseClickEvent = self.mouseClickEvent
        assert self.top_left_label is not None
        self.imageview.ui.graphicsView.addItem(self.top_left_label)

    def _init_dimensions(self, tif):
        if 'is_rgb' not in self.metadata.keys():
            self.metadata['is_rgb'] = tif.ndim == 4
        self.nDims = len(np.shape(tif))  #: int: The number of dimensions of the stored image.
        dimensions_txt = ""
        if self.nDims == 3:
            if self.metadata['is_rgb']:
                self.mx, self.my, mc = tif.shape
                self.mt = 1
                dimensions_txt = "{}x{} pixels; {} colors; ".format(self.mx, self.my, mc)
            else:
                self.mt, self.mx, self.my = tif.shape
                dimensions_txt = "{} frames; {}x{} pixels; ".format(self.mt, self.mx, self.my)
        elif self.nDims == 4:
            self.mt, self.mx ,self.my, mc = tif.shape
            dimensions_txt = "{} frames; {}x{} pixels; {} colors; ".format(self.mt, self.mx, self.my, mc)
        elif self.nDims == 2:
            self.mt = 1
            self.mx, self.my = tif.shape
            dimensions_txt = "{}x{} pixels; ".format(self.mx, self.my)
        dimensions_txt += 'dtype=' + str(self.dtype)
        if self.framerate is None:
            if 'timestamps' in self.metadata:
                ts = self.metadata['timestamps']
                self.framerate = (ts[-1] - ts[0]) / len(ts)
        if self.framerate is not None:
            dimensions_txt += '; {:.4f} {}/frame'.format(self.framerate, self.metadata['timestamp_units'])

        if self.top_left_label is not None and self.imageview is not None and self.top_left_label in self.imageview.ui.graphicsView.items():
            self.imageview.ui.graphicsView.removeItem(self.top_left_label)
        self.top_left_label = pg.LabelItem(dimensions_txt, justify='right')

    def _init_scatterplot(self):
        if self.scatterPlot in self.imageview.ui.graphicsView.items():
            self.imageview.ui.graphicsView.removeItem(self.scatterPlot)
        pointSize = g.settings['point_size']
        pointColor = QtGui.QColor(g.settings['point_color'])
        self.scatterPlot = pg.ScatterPlotItem(size=pointSize, pen=pg.mkPen([0, 0, 0, 255]), brush=pg.mkBrush(*pointColor.getRgb()))  #this is the plot that all the red points will be drawn on
        self.scatterPoints = [[] for _ in np.arange(self.mt)]
        self.scatterPlot.sigClicked.connect(self.clickedScatter)
        self.imageview.addItem(self.scatterPlot)

    def _init_menu(self):
        self.menu = QtWidgets.QMenu(self)

        def updateMenu():
            from .roi import ROI_Base
            pasteAct.setEnabled(isinstance(g.clipboard, (list, ROI_Base)))

        pasteAct = QtWidgets.QAction("&Paste", self, triggered=self.paste)
        plotAllAct = QtWidgets.QAction('&Plot All ROIs', self.menu, triggered=self.plotAllROIs )
        copyAll = QtWidgets.QAction("Copy All ROIs", self.menu, triggered = lambda a: setattr(g, 'clipboard', self.rois))
        removeAll = QtWidgets.QAction("Remove All ROIs", self.menu, triggered = self.removeAllROIs)
        self.menu.addAction(pasteAct)
        self.menu.addAction(plotAllAct)
        self.menu.addAction(copyAll)
        self.menu.addAction(removeAll)
        self.menu.aboutToShow.connect(updateMenu)

    def onResize(self, event):
        g.settings['window_settings']['coords'] = self.geometry().getRect()

    def onMove(self, event):
        g.settings['window_settings']['coords'] = self.geometry().getRect()

    def save(self, filename):
        """save(self, filename)
        Saves the current window to a specificed directory as a (.tif) file

        Args:
            filename (str): The filename, including the full path, where this (.tif) file will be saved.

        """
        from .process.file_ import save_file
        old_curr_win = g.win
        self.setAsCurrentWindow()
        save_file(filename)
        old_curr_win.setAsCurrentWindow()

    def normLUT(self, tif):
        if self.nDims == 2:
            # if the image is binary (either all 0s or 0s and 1s)
            if np.min(tif) == 0 and (np.max(tif) == 0 or np.max(tif) == 1):
                self.imageview.setLevels(-.01, 1.01)  # set levels from slightly below 0 to 1
            else:
                r = (np.min(tif), np.max(tif))  # set the levels to be just above and below the min and max of the image
                r = (r[0] - (r[1] - r[0]) / 100, r[1] + (r[1] - r[0]) / 100)
                self.imageview.setLevels(r[0], r[1])
        if self.nDims == 3 and not self.metadata['is_rgb']:
            if np.all(tif[self.currentIndex] == 0):  # if the current frame is all zeros
                r = (np.min(tif), np.max(tif))  # set the levels to be just above and below the min and max of the entire tif
                r = (r[0] - (r[1] - r[0]) / 100, r[1] + (r[1] - r[0]) / 100)
                self.imageview.setLevels(r[0], r[1])
            else:
                r = (np.min(tif[self.currentIndex]),
                     np.max(tif[self.currentIndex]))  # set the levels to be just above and below the min and max of the first frame
                r = (r[0] - (r[1] - r[0]) / 100, r[1] + (r[1] - r[0]) / 100)
                self.imageview.setLevels(r[0], r[1])
        elif self.nDims == 4 and not self.metadata['is_rgb']:
            if np.min(tif) == 0 and (np.max(tif) == 0 or np.max(tif) == 1):  # if the image is binary (either all 0s or 0s and 1s)
                self.imageview.setLevels(-.01, 1.01)  # set levels from slightly below 0 to 1

    def set_bg_im(self):
        self.bg_im_dialog = Bg_im_dialog(self)
        self.bg_im_dialog.show()

    def link(self, win):
        """link(self, win)
        Linking a window to another means when the current index of one changes, the index of the other will automatically change.

        Args:
            win (flika.window.Window): The window that will be linked with this one
        """
        if win not in self.linkedWindows:
            self.sigTimeChanged.connect(win.imageview.setCurrentIndex)
            self.linkedWindows.add(win)
            win.link(self)

    def unlink(self, win):
        """unlink(self, win)
        This unlinks a window from this one.

        Args:
            win (flika.window.Window): The window that will be unlinked from this one
        """
        if win in self.linkedWindows:
            self.linkedWindows.remove(win)
            self.sigTimeChanged.disconnect(win.imageview.setCurrentIndex)
            win.unlink(self)

    def link_toggled(self, win):
        return lambda b: self.link(win) if b else self.unlink(win)

    def make_link_menu(self):
        self.linkMenu.clear()
        for win in g.windows:
            if win == self or not win.isVisible():
                continue
            win_action = QtWidgets.QAction("%s" % win.name, self.linkMenu, checkable=True)
            win_action.setChecked(win in self.linkedWindows)
            win_action.toggled.connect(self.link_toggled(win))
            self.linkMenu.addAction(win_action)
        
    def updateindex(self):
        if self.mt == 1:
            t = 0
        else:
            (idx, t) = self.imageview.timeIndex(self.imageview.timeLine)
            t = int(np.floor(t))
        if 0 <= t < self.mt:
            self.currentIndex = t
            if not g.settings['show_all_points']:
                pointSizes = [pt[3] for pt in self.scatterPoints[t]]
                brushes = [pg.mkBrush(*pt[2].getRgb()) for pt in self.scatterPoints[t]]
                self.scatterPlot.setPoints(pos=self.scatterPoints[t], size=pointSizes, brush=brushes)
            self.sigTimeChanged.emit(t)

    def setIndex(self, index):
        """setIndex(self, index)
        This sets the index (frame) of this window. 

        Args:
            index (int): The index of the image this window will display
        """
        if hasattr(self, 'image') and self.image.ndim > 2 and 0 <= index < len(self.image):
            self.imageview.setCurrentIndex(index)

    def showFrame(self, index):
        if index>=0 and index<self.mt:
            msg = 'frame {}'.format(index)
            if 'timestamps' in self.metadata and self.metadata['timestamp_units']=='ms':
                ttime = self.metadata['timestamps'][index]
                if ttime < 1*1000:
                    msg += '; {:.4f} ms'.format(ttime)
                elif ttime < 60*1000:
                    seconds = ttime / 1000
                    msg += '; {:.4f} s'.format(seconds)
                elif ttime < 3600*1000:
                    minutes = int(np.floor(ttime / (60*1000)))
                    seconds = (ttime/1000) % 60
                    msg += '; {} m {:.4f} s'.format(minutes, seconds)
                else:
                    seconds = ttime/1000
                    hours = int(np.floor(seconds / 3600))
                    mminutes = seconds - hours * 3600
                    minutes = int(np.floor(mminutes / 60))
                    seconds = mminutes - minutes * 60
                    msg += '; {} h {} m {:.4f} s'.format(hours, minutes, seconds)
            g.m.statusBar().showMessage(msg)

    def setName(self,name):
        """setName(self,name)
        Set the name of this window.

        Args:
            name (str): the name for window to be set to
        """
        name = str(name)
        self.name = name
        self.setWindowTitle(name)
        
    def reset(self):
        if not self.closed:
            currentIndex = int(self.currentIndex)
            self.imageview.setImage(self.image, autoLevels=True) #I had autoLevels=False before.  I changed it to adjust after boolean previews.
            if self.imageview.axes['t'] is not None:
                self.imageview.setCurrentIndex(currentIndex)
            g.m.statusBar().showMessage('')

    def closeEvent(self, event):
        if self.closed:
            print('Attempt to close window {} that was already closed'.format(self))
            event.accept()
        else:
            self.closeSignal.emit()
            for win in list(self.linkedWindows):
                self.unlink(win)
            if hasattr(self,'image'):
                del self.image
            self.imageview.setImage(np.zeros((2,2))) #clear the memory
            self.imageview.close()
            del self.imageview
            if g.win==self:
                g.win=None
            if self in g.windows:
                g.windows.remove(self)
            self.closed=True
            event.accept() # let the window close

    def imageArray(self):
        """imageArray(self)

        Returns:
             Image as a 3d array, correcting for color or 2d image
        """
        tif = self.image
        nDims = len(tif.shape)
        if nDims == 4:  # If this is an RGB image stack  #[t, x, y, colors]
            tif = np.mean(tif,3)
            mx, my = tif[0, :, :].shape
        elif nDims == 3:
            if self.metadata['is_rgb']:  # [x, y, colors]
                tif = np.mean(tif,2)
                mx, my = tif.shape
                tif = tif[np.newaxis]
            else: 
                mx, my = tif[0,:,:].shape
        elif nDims == 2:
            mx, my = tif.shape
            tif = tif[np.newaxis]
        return tif

    def imageDimensions(self):
        nDims = self.image.shape
        if len(nDims) == 4: #if this is an RGB image stack
            return nDims[1:3]
        elif len(nDims) == 3:
            if self.metadata['is_rgb']:  # [x, y, colors]
                return nDims[:2]
            else:                               # [t, x, y]
                return nDims[1:]
        if len(nDims) == 2:  # If this is a static image
            return nDims
        return nDims

    def resizeEvent(self, event):
        event.accept()
        self.imageview.resize(self.size())

    def paste(self):
        """ paste(self)
        This function pastes a ROI from one window into another.
        The ROIs will be automatically linked using the link() fucntion so that when you alter one of them, the other will be altered in the same way.
        """
        def pasteROI(roi):
            if roi in self.rois:
                return None
            if roi.kind == 'rect_line':
                self.currentROI = makeROI(roi.kind, roi.pts, self, width=roi.width)
            else:
                self.currentROI = makeROI(roi.kind, roi.pts, self)
            if roi in roi.window.rois:
                self.currentROI.link(roi)
            return self.currentROI

        if type(g.clipboard) == list:
            rois = []
            for roi in g.clipboard:
                rois.append(pasteROI(roi))
            return rois
        else:
            return pasteROI(g.clipboard)
        
    def mousePressEvent(self,ev):
        ev.accept()
        self.setAsCurrentWindow()

    def setAsCurrentWindow(self):
        """setAsCurrentWindow(self)
        This function sets this window as the current window. There is only one current window. All operations are performed on the
        current window. The current window can be accessed from the variable ``g.win``. 
        """

        if g.win is not None:
            g.win.setStyleSheet("border:1px solid rgb(0, 0, 0); ")
            g.win.lostFocusSignal.emit()
        g.win = self
        g.m.currentWindow = g.win
        g.currentWindow = g.win
        if self not in g.windows:
            g.windows.append(self)
        g.m.setWindowTitle("flika - {}".format(self.name))
        self.setStyleSheet("border:1px solid rgb(0, 255, 0); ")
        g.m.setCurrentWindowSignal.sig.emit()
        self.gainedFocusSignal.emit()
    
    def clickedScatter(self, plot, points):
        p = points[0]
        x, y = p.pos()

        if g.settings['show_all_points']:
            pts = []
            for t in np.arange(self.mt):
                self.scatterPoints[t] = [p for p in self.scatterPoints[t] if not (x == p[0] and y == p[1])]
                pts.extend(self.scatterPoints[t])
            pointSizes = [pt[3] for pt in pts]
            brushes = [pg.mkBrush(*pt[2].getRgb()) for pt in pts]
            self.scatterPlot.setPoints(pos=pts, size=pointSizes, brush=brushes)
        else:
            t = self.currentIndex
            self.scatterPoints[t] = [p for p in self.scatterPoints[t] if not (x == p[0] and y == p[1])]
            pointSizes = [pt[3] for pt in self.scatterPoints[t]]
            brushes = [pg.mkBrush(*pt[2].getRgb()) for pt in self.scatterPoints[t]]
            self.scatterPlot.setPoints(pos=self.scatterPoints[t], size=pointSizes, brush=brushes)

    def getScatterPts(self):
        """getScatterPts(self)

        Returns:
            numpy array: an Nx3 array of scatter points, where N is the number of points. Col0 is frame, Col1 is x, Col2 is y. 
        """
        p_out=[]
        p_in=self.scatterPoints
        for t in np.arange(len(p_in)):
            for p in p_in[t]:
                p_out.append(np.array([t,p[0],p[1]]))
        p_out=np.array(p_out)
        return p_out

    def plotAllROIs(self):
        for roi in self.rois:
            if roi.traceWindow == None:
                roi.plot()

    def removeAllROIs(self):
        for roi in self.rois[:]:
            roi.delete()

    def addPoint(self, p=None):
        if p is None:
            p = [self.currentIndex, self.x, self.y]
        elif len(p) != 3:
            raise Exception("addPoint takes a 3-tuple (t, x, y) as argument")

        t, x, y = p

        pointSize=g.m.settings['point_size']
        pointColor = QtGui.QColor(g.settings['point_color'])
        position=[x, y, pointColor, pointSize]
        self.scatterPoints[t].append(position)
        self.scatterPlot.addPoints(pos=[[x, y]], size=pointSize, brush=pg.mkBrush(*pointColor.getRgb()))

    def mouseClickEvent(self,ev):
        ''''mouseClickevent(self, ev)
        Event handler for when the mouse is pressed in a flika window.
        '''
        self.EEEE = ev
        if self.x is not None and self.y is not None and ev.button() == 2 and not self.creatingROI:
            mm = g.settings['mousemode']
            if mm == 'point':
                self.addPoint()
            elif mm == 'rectangle' and g.settings['default_roi_on_click']:
                pts = [pg.Point(self.x - g.settings['rect_width']/2, self.y - g.settings['rect_height']/2), pg.Point(g.settings['rect_width'], g.settings['rect_height'])]
                self.currentROI = makeROI("rectangle", pts)
            else:
                self.menu.exec_(ev.screenPos().toQPoint())
        elif self.creatingROI:
            self.currentROI.cancel()
            self.creatingROI = None

    def save_rois(self, filename=None):
        """save_rois(self, filename=None)

        Args:
            filename (str): The filename, including the full path, where the ROI file will be saved.

        """
        if not isinstance(filename, str):
            if filename is not None and os.path.isfile(filename):
                filename = os.path.splitext(g.settings['filename'])[0]
                filename = save_file_gui('Save ROI', filename, '*.txt')
            else:
                filename = save_file_gui('Save ROI', '', '*.txt')

        if filename != '' and isinstance(filename, str):
            reprs = [roi._str() for roi in self.rois]
            reprs = '\n'.join(reprs)
            open(filename, 'w').write(reprs)
        else:
            g.m.statusBar().showMessage('No File Selected')

    
    def keyPressEvent(self, ev):
        if ev.key() == QtCore.Qt.Key_Delete:
            i = 0
            while i < len(self.rois):
                if self.rois[i].mouseHovering:
                    self.rois[i].delete()
                else:
                    i += 1
        self.keyPressSignal.emit(ev)
        
    def mouseMoved(self,point):
        '''mouseMoved(self,point)
        Event handler function for mouse movement.
        '''
        point=self.imageview.getImageItem().mapFromScene(point)
        self.point = point
        self.x = point.x()
        self.y = point.y()
        image=self.imageview.getImageItem().image
        if self.x < 0 or self.y < 0 or self.x >= image.shape[0] or self.y >= image.shape[1]:
            pass# if we are outside the image
        else:
            z=self.imageview.currentIndex
            value=image[int(self.x),int(self.y)]
            g.m.statusBar().showMessage('x={}, y={}, z={}, value={}'.format(int(self.x),int(self.y),z,value))

    def mouseDragEvent(self, ev):
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ShiftModifier:
            pass #This is how I detect that the shift key is held down.
        if ev.button() == QtCore.Qt.LeftButton:
            ev.accept()
            if g.settings['mousemode'] == 'pencil':
                if ev.isStart():
                    self.last_x = None
                    self.last_y = None
                else:
                    if self.x < 0 or self.y < 0 or self.x >= self.mx or self.y >= self.my:
                        self.last_x = None  # if we are outside the image
                        self.last_y = None
                    else:
                        x = int(self.x)
                        y = int(self.y)
                        if self.last_x == x and self.last_y == y:
                            pass
                        z = self.imageview.currentIndex
                        image = self.imageview.getImageItem().image
                        v = g.settings['pencil_value']
                        if self.last_x is None:
                            image[x, y] = v
                        else:
                            xs, ys = get_line(x, y, self.last_x, self.last_y).T
                            image[xs, ys] = v
                        self.imageview.imageItem.updateImage(image)
                        self.last_x = x
                        self.last_y = y
                        self.ev = ev
            else:
                difference=self.imageview.getImageItem().mapFromScene(ev.lastScenePos())-self.imageview.getImageItem().mapFromScene(ev.scenePos())
                self.imageview.view.translateBy(difference)
        if ev.button() == QtCore.Qt.RightButton:
            ev.accept()
            mm = g.settings['mousemode']
            if mm in ('freehand', 'line', 'rectangle', 'rect_line'):
                if ev.isStart():
                    self.ev = ev
                    pt = self.imageview.getImageItem().mapFromScene(ev.buttonDownScenePos())
                    self.x = pt.x()  # This sets x and y to the button down position, not the current position.
                    self.y = pt.y()
                    self.creatingROI = True
                    self.currentROI = ROI_Drawing(self, self.x, self.y, mm)
                if ev.isFinish():
                    if self.creatingROI:   
                        if ev._buttons | QtCore.Qt.RightButton != ev._buttons:
                            self.currentROI = self.currentROI.drawFinished()
                            self.creatingROI = False
                        else:
                            self.currentROI.cancel()
                            self.creatingROI = False
                else:  # If we are in the middle of the drag between starting and finishing.
                    if self.creatingROI:
                        self.currentROI.extend(self.x, self.y)

    def updateTimeStampLabel(self,frame):
        label = self.timeStampLabel
        if self.framerate == 0:
            label.setHtml("<span style='font-size: 12pt;color:white;background-color:None;'>Frame rate is 0 Hz</span>" )
            return False
        ttime = frame/self.framerate  # Time elapsed since the first frame until the current frame, in seconds.
        if ttime < 1:
            ttime = ttime * 1000
            label.setHtml("<span style='font-size: 12pt;color:white;background-color:None;'>{:.0f} ms</span>".format(ttime))
        elif ttime < 60:
            label.setHtml("<span style='font-size: 12pt;color:white;background-color:None;'>{:.3f} s</span>".format(ttime))
        elif ttime < 3600:
            minutes = int(np.floor(ttime/60))
            seconds = ttime % 60
            label.setHtml("<span style='font-size: 12pt;color:white;background-color:None;'>{}m {:.3f} s</span>".format(minutes,seconds))
        else:
            hours = int(np.floor(ttime/3600))
            mminutes = ttime-hours*3600
            minutes = int(np.floor(mminutes/60))
            seconds = mminutes-minutes*60
            label.setHtml("<span style='font-size: 12pt;color:white;background-color:None;'>{}h {}m {:.3f} s</span>".format(hours,minutes,seconds))
logger.debug("Completed 'reading window.py'")


def get_line(x1, y1, x2, y2):
    """Bresenham's Line Algorithm
    Produces a list of tuples """
    # Setup initial conditions
    dx = x2 - x1
    dy = y2 - y1

    # Determine how steep the line is
    is_steep = abs(dy) > abs(dx)

    # Rotate line
    if is_steep:
        x1, y1 = y1, x1
        x2, y2 = y2, x2

    # Swap start and end points if necessary and store swap state
    swapped = False
    if x1 > x2:
        x1, x2 = x2, x1
        y1, y2 = y2, y1
        swapped = True

    # Recalculate differentials
    dx = x2 - x1
    dy = y2 - y1

    # Calculate error
    error = int(dx / 2.0)
    ystep = 1 if y1 < y2 else -1

    # Iterate over bounding box generating points between start and end
    y = y1
    points = []
    for x in range(x1, x2 + 1):
        coord = [y, x] if is_steep else [x, y]
        points.append(coord)
        error -= abs(dy)
        if error < 0:
            y += ystep
            error += dx

    # Reverse the list if the coordinates were swapped
    if swapped:
        points.reverse()
    return np.array(points)


# -*- coding: utf-8 -*-
from ..logger import logger
logger.debug("Started 'reading process/file_.py'")
import pyqtgraph as pg
import pyqtgraph.exporters
import time
import os.path
import numpy as np
from qtpy import uic, QtGui, QtCore, QtWidgets
import shutil, subprocess
import datetime
import json
import re
import pathlib

from .. import global_vars as g
from ..utils.BaseProcess import BaseDialog
from ..window import Window
from ..utils.misc import open_file_gui, save_file_gui
from ..utils.io import tifffile

__all__ = ['save_file', 'save_points', 'save_rois', 'save_movie_gui', 'open_file', 'open_file_from_gui', 'open_image_sequence_from_gui', 'open_points', 'close']

########################################################################################################################
######################                  SAVING FILES                                         ###########################
########################################################################################################################



def save_file(filename=None):
    """save_file(filename=None)
    Save the image in the currentWindow to a .tif file.

    Parameters:
        filename (str): The image or movie will be saved as  'filename'.tif.

    """
    if filename is None or filename is False:
        filetypes = '*.tif'
        prompt = 'Save File As Tif'
        filename = save_file_gui(prompt, filetypes=filetypes)
        if filename is None:
            return None
    if os.path.dirname(filename) == '':  # if the user didn't specify a directory
        directory = os.path.normpath(os.path.dirname(g.settings['filename']))
        filename = os.path.join(directory, filename)
    g.m.statusBar().showMessage('Saving {}'.format(os.path.basename(filename)))
    A = g.win.image
    if A.dtype == np.bool:
        A = A.astype(np.uint8)
    metadata = g.win.metadata
    try:
        metadata = json.dumps(metadata, default=JSONhandler)
    except TypeError as e:
        msg = "Error saving metadata.\n{}\nContinuing to save file".format(e)
        g.alert(msg)
    if len(A.shape) == 3:
        A = np.transpose(A, (0, 2, 1))  # This keeps the x and the y the same as in FIJI
    elif len(A.shape) == 2:
        A = np.transpose(A, (1, 0))
    tifffile.imsave(filename, A,
                    description=metadata)  # http://stackoverflow.com/questions/20529187/what-is-the-best-way-to-save-image-metadata-alongside-a-tif-with-python
    g.m.statusBar().showMessage('Successfully saved {}'.format(os.path.basename(filename)))
    return filename

def save_points(filename=None):
    """save_points(filename=None)
    Saves the points in the current window to a text file

    Parameters:
        filename (str): Address to save the points to, with .txt


    """

    if filename is None:
        filetypes = '*.txt'
        prompt = 'Save Points'
        filename = save_file_gui(prompt, filetypes=filetypes)
        if filename is None:
            return None
    g.m.statusBar().showMessage('Saving Points in {}'.format(os.path.basename(filename)))
    p_out = []
    p_in = g.win.scatterPoints
    for t in np.arange(len(p_in)):
        for p in p_in[t]:
            p_out.append(np.array([t, p[0], p[1]]))
    p_out = np.array(p_out)
    np.savetxt(filename, p_out)
    g.m.statusBar().showMessage('Successfully saved {}'.format(os.path.basename(filename)))
    return filename


def save_movie_gui():
    rateSpin = pg.SpinBox(value=50, bounds=[1, 1000], suffix='fps', int=True, step=1)
    rateDialog = BaseDialog([{'string': 'Framerate', 'object': rateSpin}], 'Save Movie', 'Set the framerate')
    rateDialog.accepted.connect(lambda: save_movie(rateSpin.value()))
    g.dialogs.append(rateDialog)
    rateDialog.show()

def save_rois( filename=None):
    g.currentWindow.save_rois(filename)


def save_movie(rate, filename=None):
    """save_movie(rate, filename)
    Saves the currentWindow video as a .mp4 movie by joining .jpg frames together

    Parameters:
        rate (int): framerate
        filename (str): Address to save the movie to, with .mp4

    Notes:
        Once you've exported all of the frames you wanted, open a command line and run the following:
        ffmpeg -r 100 -i %03d.jpg output.mp4
        -r: framerate
        -i: input files.
        %03d: The files have to be numbered 001.jpg, 002.jpg... etc.

    """


    ## Check if ffmpeg is installed
    if os.name == 'nt':  # If we are running windows
        try:
            subprocess.call(["ffmpeg"])
        except FileNotFoundError as e:
            if e.errno == os.errno.ENOENT:
                # handle file not found error.
                # I used http://ffmpeg.org/releases/ffmpeg-2.8.4.tar.bz2 originally
                g.alert("The program FFmpeg is required to export movies. \
                \n\nFor instructions on how to install, go here: http://www.wikihow.com/Install-FFmpeg-on-Windows")
                return None
            else:
                # Something else went wrong while trying to run `wget`
                raise

    filetypes = "Movies (*.mp4)"
    prompt = "Save movie to .mp4 file"
    filename = save_file_gui(prompt, filetypes=filetypes)
    if filename is None:
        return None

    win = g.win
    A = win.image
    if len(A.shape) < 3:
        g.alert('Movie not the right shape for saving.')
        return None
    try:
        exporter = pg.exporters.ImageExporter(win.imageview.view)
    except TypeError:
        exporter = pg.exporters.ImageExporter.ImageExporter(win.imageview.view)

    nFrames = len(A)
    tmpdir = os.path.join(os.path.dirname(g.settings.settings_file), 'tmp')
    if os.path.isdir(tmpdir):
        shutil.rmtree(tmpdir)
    os.mkdir(tmpdir)
    win.top_left_label.hide()
    for i in np.arange(0, nFrames):
        win.setIndex(i)
        exporter.export(os.path.join(tmpdir, '{:03}.jpg'.format(i)))
        QtWidgets.qApp.processEvents()
    win.top_left_label.show()
    olddir = os.getcwd()
    os.chdir(tmpdir)
    subprocess.call(
        ['ffmpeg', '-r', '%d' % rate, '-i', '%03d.jpg', '-vf', 'scale=trunc(iw/2)*2:trunc(ih/2)*2', 'output.mp4'])
    os.rename('output.mp4', filename)
    os.chdir(olddir)
    g.m.statusBar().showMessage('Successfully saved movie as {}.'.format(os.path.basename(filename)))










########################################################################################################################
######################                         OPENING FILES                                 ###########################
########################################################################################################################

def open_image_sequence_from_gui():
    open_image_sequence(None, True)

def open_file_from_gui():
    open_file(None, True)


def open_image_sequence(filename=None, from_gui=False):
    """ open_image_sequencefilename(filename=None)
    Opens an image sequence (.tif, .png) into a newWindow.

    Parameters:
        filename (str): Address of the first of a series of files that will be stitched together into a movie.
                            If no filename is provided, the last opened file is used.
    Returns:
        newWindow

    """
    if filename is None:
        if from_gui:
            filetypes = 'Image Files (*.tif *.tiff *.png);;All Files (*.*)'
            prompt = 'Open File'
            filename = open_file_gui(prompt, filetypes=filetypes)
            if filename is None:
                return None
        else:
            filename = g.settings['filename']
            if filename is None:
                g.alert('No filename selected')
                return None
    print("Filename: {}".format(filename))
    g.m.statusBar().showMessage('Loading {}'.format(os.path.basename(filename)))
    t = time.time()
    metadata = dict()

    filename = pathlib.Path(filename)
    assert filename.is_file()
    directory = filename.parents[0]
    ext = filename.suffix
    all_image_filenames = [p for p in directory.iterdir() if p.suffix == ext]

    all_images = []
    if ext in ['.tif', '.stk', '.tiff', '.ome']:
        for f in all_image_filenames:
            results = open_tiff(f, metadata)
            if results is None:
                return None
            else:
                A, metadata = results
                all_images.append(A)
        all_images = np.array(all_images)
    elif ext in ['.png']:
        pass

    append_recent_file(str(filename))  # make first in recent file menu
    g.m.statusBar().showMessage('{} successfully loaded ({} s)'.format(filename.parts[-1], time.time() - t))
    g.settings['filename'] = str(filename)
    commands = ["open_image_sequence('{}')".format(str(filename))]
    newWindow = Window(all_images, filename.parts[-1], str(filename), commands, metadata)
    return newWindow


def open_file(filename=None, from_gui=False):
    """ open_file(filename=None)
    Opens an image or movie file (.tif, .stk, .nd2) into a newWindow.

    Parameters:
        filename (str): Address of file to open. If no filename is provided, the last opened file is used.
    Returns:
        newWindow

    """
    if filename is None:
        if from_gui:
            filetypes = 'Image Files (*.tif *.stk *.tiff *.nd2);;All Files (*.*)'
            prompt = 'Open File'
            filename = open_file_gui(prompt, filetypes=filetypes)
            if filename is None:
                return None
        else:
            filename = g.settings['filename']
            if filename is None:
                g.alert('No filename selected')
                return None
    print("Filename: {}".format(filename))
    g.m.statusBar().showMessage('Loading {}'.format(os.path.basename(str(filename))))
    t = time.time()
    metadata = dict()
    ext = os.path.splitext(str(filename))[1]
    if ext in ['.tif', '.stk', '.tiff', '.ome']:
        results = open_tiff(str(filename), metadata)
        if results is None:
            return None
        else:
            A, metadata = results
    elif ext == '.nd2':
        import nd2reader
        nd2 = nd2reader.ND2Reader(str(filename))
        axes = nd2.axes
        mx = nd2.metadata['width']
        my = nd2.metadata['height']
        mt = nd2.metadata['total_images_per_channel']
        A = np.zeros((mt, mx, my))
        percent = 0
        for frame in range(mt):
            A[frame] = nd2[frame].T
            if percent < int(100 * float(frame) / mt):
                percent = int(100 * float(frame) / mt)
                g.m.statusBar().showMessage('Loading file {}%'.format(percent))
                QtWidgets.qApp.processEvents()
        metadata = nd2.metadata
    elif ext == '.py':
        from ..app.script_editor import ScriptEditor
        ScriptEditor.importScript(filename)
        return
    elif ext == '.whl':
        import pip
        # first, remove trailing (1) or (2)
        newfilename = re.sub(r' \([^)]*\)', '', filename)
        try:
            os.rename(filename, newfilename)
        except FileExistsError:
            pass
        filename = newfilename
        result = pip.main(['install', filename])
        if result == 0:
            g.alert('Successfully installed {}'.format(filename))
        else:
            g.alert('Install of {} failed'.format(filename))
        return
    elif ext == '.jpg' or ext == '.png':
        import skimage.io
        A = skimage.io.imread(filename)
        if len(A.shape) == 3:
            perm = get_permutation_tuple(['y', 'x', 'c'], ['x', 'y', 'c'])
            A = np.transpose(A, perm)
            metadata['is_rgb'] = True

    else:
        msg = "Could not open.  Filetype for '{}' not recognized".format(filename)
        g.alert(msg)
        if filename in g.settings['recent_files']:
            g.settings['recent_files'].remove(filename)
        # make_recent_menu()
        return
        
    append_recent_file(str(filename))  # make first in recent file menu
    g.m.statusBar().showMessage('{} successfully loaded ({} s)'.format(os.path.basename(str(filename)), time.time() - t))
    g.settings['filename'] = str(filename)
    commands = ["open_file('{}')".format(filename)]
    newWindow = Window(A, os.path.basename(str(filename)), filename, commands, metadata)
    return newWindow

def open_tiff(filename, metadata):
    try:
        Tiff = tifffile.TiffFile(str(filename))
    except Exception as s:
        g.alert("Unable to open {}. {}".format(filename, s))
        return None
    metadata = get_metadata_tiff(Tiff)
    A = Tiff.asarray()
    Tiff.close()
    axes = [tifffile.AXES_LABELS[ax] for ax in Tiff.series[0].axes]
    # print("Original Axes = {}".format(Tiff.series[0].axes)) #sample means RBGA, plane means frame, width means X, height means Y
    try:
        assert len(axes) == len(A.shape)
    except AssertionError:
        msg = 'Tiff could not be loaded because the number of axes in the array does not match the number of axes found by tifffile.py\n'
        msg += "Shape of array: {}\n".format(A.shape)
        msg += "Axes found by tifffile.py: {}\n".format(axes)
        g.alert(msg)
        return None
    if set(axes) == set(['height', 'width']):  # still image in black and white.
        target_axes = ['width', 'height']
    elif set(axes) == set(['height', 'width', 'channel']):  # still image in color.
        target_axes = ['width', 'height', 'channel']
        metadata['is_rgb'] = True
    elif set(axes) == set(['height', 'width', 'sample']):  # still image in color.
        target_axes = ['width', 'height', 'sample']
        metadata['is_rgb'] = True
    elif set(axes) == set(['height', 'width', 'series']):  # movie in black and white
        target_axes = ['series', 'width', 'height']
    elif set(axes) == set(['height', 'width', 'time']):  # movie in black and white
        target_axes = ['time', 'width', 'height']
    elif set(axes) == set(['height', 'width', 'depth']):  # movie in black and white
        target_axes = ['depth', 'width', 'height']
    elif set(axes) == set(['channel', 'time', 'height', 'width']):  # movie in color
        target_axes = ['time', 'width', 'height', 'channel']
        metadata['is_rgb'] = True
    elif set(axes) == set(['sample', 'time', 'height', 'width']):  # movie in color
        target_axes = ['time', 'width', 'height', 'sample']
        metadata['is_rgb'] = True
    elif set(axes) == set(['other', 'height', 'width']):
        target_axes = ['other', 'height', 'width']
        metadata['is_rgb'] = False

    perm = get_permutation_tuple(axes, target_axes)
    A = np.transpose(A, perm)
    if target_axes[-1] in ['channel', 'sample', 'series'] and A.shape[-1] == 2:
        B = np.zeros(A.shape[:-1])
        B = np.expand_dims(B, len(B.shape))
        A = np.append(A, B, len(A.shape) - 1)  # add a column of zeros to the last dimension.
        # if A.ndim == 4 and axes[3] == 'sample' and A.shape[3] == 1:
        #    A = np.squeeze(A)  # this gets rid of the meaningless 4th dimention in .stk files
    return [A, metadata]


def open_points(filename=None):
    """open_points(filename=None)
    Opens a specified text file and displays the points from that file into the currentWindow

    Parameters:
        filename (str): Address of file to open. If no filename is provided, the last opened file is used.

    Note:
        Any existing points on a currentWindow will persist when another points file is opened and displayed

    """
    if g.win is None:
        g.alert('Points cannot be loaded if no window is selected. Open a file and click on a window.')
        return None
    if filename is None:
        filetypes = '*.txt'
        prompt = 'Load Points'
        filename = open_file_gui(prompt, filetypes=filetypes)
        if filename is None:
            return None
    g.m.statusBar().showMessage('Loading points from {}'.format(os.path.basename(filename)))
    try:
        pts = np.loadtxt(filename)
    except UnicodeDecodeError:
        g.alert('This points file contains text that cannot be read. No points loaded.')
        return None
    if len(pts) == 0:
        g.alert('This points file is empty. No points loaded.')
        return None
    nCols = pts.shape[1]
    pointSize = g.settings['point_size']
    pointColor = QtGui.QColor(g.settings['point_color'])
    if nCols == 3:
        for pt in pts:
            t = int(pt[0])
            if g.win.mt == 1:
                t = 0
            g.win.scatterPoints[t].append([pt[1],pt[2], pointColor, pointSize])
        if g.settings['show_all_points']:
            pts = []
            for t in np.arange(g.win.mt):
                pts.extend(g.win.scatterPoints[t])
            pointSizes = [pt[3] for pt in pts]
            brushes = [pg.mkBrush(*pt[2].getRgb()) for pt in pts]
            g.win.scatterPlot.setPoints(pos=pts, size=pointSizes, brush=brushes)
        else:
            t = g.win.currentIndex
            g.win.scatterPlot.setPoints(pos=g.win.scatterPoints[t])
            g.win.updateindex()
    elif nCols == 2:
        t = 0
        for pt in pts:
            g.win.scatterPoints[t].append([pt[0], pt[1], pointColor, pointSize])
        t = g.win.currentIndex
        g.win.scatterPlot.setPoints(pos=g.win.scatterPoints[t])

    g.m.statusBar().showMessage('Successfully loaded {}'.format(os.path.basename(filename)))




    
########################################################################################################################
######################                INTERNAL HELPER FUNCTIONS                              ###########################
########################################################################################################################
def get_permutation_tuple(src, dst):
    """get_permtation_tuple(src, dst)

    Parameters:
        src (list): The original ordering of the axes in the tiff.
        dst (list): The desired ordering of the axes in the tiff.

    Returns:
        result (tuple): The required permutation so the axes are ordered as desired.
    """
    result = []
    for i in dst:
        result.append(src.index(i))
    result = tuple(result)
    return result

def append_recent_file(fname):
    fname = os.path.abspath(fname)
    if fname in g.settings['recent_files']:
        g.settings['recent_files'].remove(fname)
    if os.path.exists(fname):
        g.settings['recent_files'].append(fname)
        if len(g.settings['recent_files']) > 8:
            g.settings['recent_files'] = g.settings['recent_files'][-8:]
    return fname


def get_metadata_tiff(Tiff):
    metadata = {}
    if hasattr(Tiff[0], 'is_micromanager') and Tiff[0].is_micromanager:
        imagej_tags_unpacked = {}
        if hasattr(Tiff[0],'imagej_tags'):
            imagej_tags = Tiff[0].imagej_tags
            imagej_tags['info']
            imagej_tags_unpacked = json.loads(imagej_tags['info'])
        micromanager_metadata = Tiff[0].tags['micromanager_metadata']
        metadata = {**micromanager_metadata.value, **imagej_tags_unpacked}
        if 'Frames' in metadata and metadata['Frames'] > 1:
            timestamps = [c.tags['micromanager_metadata'].value['ElapsedTime-ms'] for c in Tiff]
            metadata['timestamps'] = timestamps
            metadata['timestamp_units'] = 'ms'
        keys_to_remove = ['NextFrame', 'ImageNumber', 'Frame', 'FrameIndex']
        for key in keys_to_remove:
            metadata.pop(key)
    else:
        try:
            metadata = Tiff[0].image_description
            metadata = txt2dict(metadata)
        except AttributeError:
            metadata = dict()
    metadata['is_rgb'] = Tiff[0].is_rgb
    return metadata



def txt2dict(metadata):
    meta = dict()
    try:
        metadata = json.loads(metadata.decode('utf-8'))
        return metadata
    except ValueError:  # if the metadata isn't in JSON
        pass
    for line in metadata.splitlines():
        line = re.split('[:=]', line.decode())
        if len(line) == 1:
            meta[line[0]] = ''
        else:
            meta[line[0].lstrip().rstrip()] = line[1].lstrip().rstrip()
    return meta


def JSONhandler(obj):
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()
    else:
        json.JSONEncoder().default(obj)


def close(windows=None):
    """close(window=None)
    Will close a window or a set of windows.

    Parameters:
        'all' (str): closes all windows
        windows (list): closes each window in the list
        Window: closes individual window
        (None): closes current window

    """
    if isinstance(windows, str):
        if windows == 'all':
            windows = [window for window in g.windows]
            for window in windows:
                window.close()
    elif isinstance(windows,list):
        for window in windows:
            if isinstance(window,Window):
                window.close()
    elif isinstance(windows,Window):
        windows.close()
    elif windows is None:
        if g.win is not None:
            g.win.close()


########################################################################################################################
######################             OLD FUNCTIONS THAT MIGHT BE USEFUL SOMEDAY                ###########################
########################################################################################################################
"""

def save_roi_traces(filename):
    g.m.statusBar().showMessage('Saving traces to {}'.format(os.path.basename(filename)))
    to_save = [roi.getTrace() for roi in g.win.rois]
    np.savetxt(filename, np.transpose(to_save), header='\t'.join(['ROI %d' % i for i in range(len(to_save))]), fmt='%.4f', delimiter='\t', comments='')
    g.settings['filename'] = filename
    g.m.statusBar().showMessage('Successfully saved traces to {}'.format(os.path.basename(filename)))

def load_metadata(filename=None):
    '''This function loads the .txt file corresponding to a file into a dictionary
    The .txt is a file which includes database connection information'''
    meta=dict()
    if filename is None:
        filename=os.path.splitext(g.settings['filename'])[0]+'.txt'
    BOM = codecs.BOM_UTF8.decode('utf8')
    if not os.path.isfile(filename):
        g.alert("'"+filename+"' is not a file.")
        return dict()
    with codecs.open(filename, encoding='utf-8') as f:
        for line in f:
            line = line.lstrip(BOM)
            line=line.split('=')
            meta[line[0].lstrip().rstrip()]=line[1].lstrip().rstrip()
    for k in meta.keys():
        if meta[k].isdigit():
            meta[k]=int(meta[k])
        else:
            try:
                meta[k]=float(meta[k])
            except ValueError:
                pass
    return meta

def save_metadata(meta):
    filename=os.path.splitext(g.settings['filename'])[0]+'.txt'
    f=open(filename, 'w')
    text=''
    for item in meta.items():
        text+="{}={}\n".format(item[0],item[1])
    f.write(text)
    f.close()


def save_current_frame(filename):
    "" save_current_frame(filename)
    Save the current single frame image of the currentWindow to a .tif file.

    Parameters:
        | filename (str) -- Address to save the frame to.
    ""
    if os.path.dirname(filename)=='': #if the user didn't specify a directory
        directory=os.path.normpath(os.path.dirname(g.settings['filename']))
        filename=os.path.join(directory,filename)
    g.m.statusBar().showMessage('Saving {}'.format(os.path.basename(filename)))
    A=np.average(g.win.image, 0)#.astype(g.settings['internal_data_type'])
    metadata=json.dumps(g.win.metadata)
    if len(A.shape)==3:
        A = A[g.win.currentIndex]
        A=np.transpose(A,(0,2,1)) # This keeps the x and the y the same as in FIJI
    elif len(A.shape)==2:
        A=np.transpose(A,(1,0))
    tifffile.imsave(filename, A, description=metadata) #http://stackoverflow.com/questions/20529187/what-is-the-best-way-to-save-image-metadata-alongside-a-tif-with-python
    g.m.statusBar().showMessage('Successfully saved {}'.format(os.path.basename(filename)))

def make_recent_menu():
    g.m.menuRecent_Files.clear()
    if len(g.settings['recent_files']) == 0:
        no_recent = QtWidgets.QAction("No Recent Files", g.m)
        no_recent.setEnabled(False)
        g.m.menuRecent_Files.addAction(no_recent)
        return
    def openFun(f):
        return lambda: open_file(append_recent_file(f))
    for fname in g.settings['recent_files'][:10]:
        if os.path.exists(fname):
            g.m.menuRecent_Files.addAction(QtWidgets.QAction(fname, g.m, triggered=openFun(fname)))

"""

logger.debug("Completed 'reading process/file_.py'")


from ..logger import logger, handle_exception
logger.debug("Started 'reading app/application.py'")

import sys, os, time
import ctypes
import platform
import traceback
logger.debug("Started 'reading app/application.py, importing qtpy'")
from qtpy import QtCore, QtWidgets, QtGui
logger.debug("Completed 'reading app/application.py, importing qtpy'")

from ..utils.misc import nonpartial
from ..utils.app import get_qapp
from ..app.settings_editor import SettingsEditor, rectSettings, pointSettings, pencilSettings
from .. import global_vars as g
from .plugin_manager import PluginManager, Load_Local_Plugins_Thread
from .script_editor import ScriptEditor
from ..utils.misc import load_ui, send_error_report, Send_User_Stats_Thread
from ..images import image_path
from ..version import __version__
from ..update_flika import checkUpdates


def status_pixmap(attention=False):
    """status_pixmap(attention=False)
    A small icon to grab attention

    Args:
        attention (bool): pixmap is red if True, gray if otherwise

    Returns:
        QtGui.QPixmap: attention icon to display
    """
    color = QtCore.Qt.red if attention else QtCore.Qt.lightGray

    pm = QtGui.QPixmap(15, 15)
    p = QtGui.QPainter(pm)
    b = QtGui.QBrush(color)
    p.fillRect(-1, -1, 20, 20, b)
    return pm


class ClickableLabel(QtWidgets.QLabel):
    """A QtGui.QLabel you can click on to generate events
    """

    clicked = QtCore.Signal()

    def mousePressEvent(self, event):
        self.clicked.emit()


class Logger(QtWidgets.QWidget):
    """A window to display error messages
    """

    def __init__(self, parent=None):
        super(Logger, self).__init__(parent)
        self._text = QtWidgets.QTextEdit()
        self._text.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        clear = QtWidgets.QPushButton("Clear")
        clear.clicked.connect(nonpartial(self._clear))
        report = QtWidgets.QPushButton("Send Bug Report")
        report.clicked.connect(nonpartial(self._send_report))
        self._status = ClickableLabel()
        self._status.setToolTip("View Errors and Warnings")
        self._status.clicked.connect(self._show)
        self._status.setPixmap(status_pixmap())
        self._status.setContentsMargins(0, 0, 0, 0)

        l = QtWidgets.QVBoxLayout()
        h = QtWidgets.QHBoxLayout()
        l.setContentsMargins(2, 2, 2, 2)
        l.setSpacing(2)
        h.setContentsMargins(0, 0, 0, 0)

        l.addWidget(self._text)
        h.insertStretch(0)
        h.addWidget(report)
        h.addWidget(clear)
        l.addLayout(h)

        self.setLayout(l)

    @property
    def status_light(self):
        """status_light(self)

        Returns:
            The icon representing the status of the log
        """
        return self._status

    def write(self, message):
        """write(self, message)
        Interface for sys.excepthook
        """
        self._text.insertPlainText(message)
        self._status.setPixmap(status_pixmap(attention=True))

    def flush(self):
        """flush(self)
        Interface for sys.excepthook
        """
        pass

    def _send_report(self):
        """_send_report(self)
        Send the contents of the log as a bug report
        """
        text = self._text.document().toPlainText()
        email = QtWidgets.QInputDialog.getText(self, "Response email", "Enter your email if you would like us to contact you about this bug.")
        if isinstance(email, tuple) and len(email) == 2:
            email = email[0]
        response = send_error_report(email=email, report=text)
        if response is None or response.status_code != 200:
            g.alert("Failed to send error report. Response {}:\n{}".format((response.status_code, response._content)))
        else:
            if email != '':
                g.alert("Bug report sent. We will contact you as soon as we can.")
            else:
                g.alert("Bug report sent. Thank you!")

    def _clear(self):
        """_clear(self)
        Erase the log
        """
        self._text.setText('')
        self._status.setPixmap(status_pixmap(attention=False))
        self.close()

    def _show(self):
        """_show(self)
        Show the log
        """
        self.show()
        self.raise_()

    def keyPressEvent(self, event):
        """keyPressEvent(self, event)
        Hide window on escape key
        """
        if event.key() == QtCore.Qt.Key_Escape:
            self.hide()



class FlikaApplication(QtWidgets.QMainWindow):
    """The main window of flika, stored as g.m
    """
    def __init__(self):
        logger.debug("Started 'creating app.application.FlikaApplication'")
        from ..process.file_ import open_file, open_file_from_gui, open_image_sequence_from_gui, open_points, save_file, save_movie_gui, save_points, save_rois
        from ..process import setup_menus
        logger.debug("Started 'creating app.application.FlikaApplication.app'")
        self.app = get_qapp(image_path('favicon.png'))
        logger.debug("Completed 'creating app.application.FlikaApplication.app'")
        super(FlikaApplication, self).__init__()
        self.app.setQuitOnLastWindowClosed(True)
        setup_menus()
        logger.debug("Started 'loading main.ui'")
        load_ui('main.ui', self, directory=os.path.dirname(__file__))
        logger.debug("Completed 'loading main.ui'")

        g.m = self
        # These are all added for backwards compatibility for plugins
        self.windows = g.windows
        self.traceWindows = g.traceWindows
        self.dialogs = g.dialogs
        self.currentWindow = g.win
        self.currentTrace = g.currentTrace
        self.clipboard = g.clipboard
        self.setWindowSize()
        if platform.system() == 'Windows':
            myappid = 'flika-org.flika.' + str(__version__)
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
        self.menuBar().setNativeMenuBar(False)
        self._make_menu()
        self._make_tools()

        self._log = Logger()
        def handle_exception_wrapper(exc_type, exc_value, exc_traceback):
            handle_exception(exc_type, exc_value, exc_traceback)
            tb_str = traceback.format_exception(exc_type, exc_value, exc_traceback)
            tb_str = ''.join(tb_str)+'\n'
            self._log.write(tb_str)
        sys.excepthook = handle_exception_wrapper

        g.dialogs.append(self._log)
        self._log.window().setWindowTitle("Console Log")
        self._log.resize(550, 550)

        self.statusBar().addPermanentWidget(self._log.status_light)
        self.statusBar().setContentsMargins(2, 0, 20, 2)
        self.statusBar().setSizeGripEnabled(False)
        self.setCurrentWindowSignal = SetCurrentWindowSignal(self)
        self.setAcceptDrops(True)
        self.load_local_plugins_thread = Load_Local_Plugins_Thread()
        self.load_local_plugins_thread.start()
        self.load_local_plugins_thread.plugins_done_sig.connect(self.plugins_done)
        self.load_local_plugins_thread.error_loading.connect(g.alert)
        logger.debug("Completed 'creating app.application.FlikaApplication'")

    def plugins_done(self, plugins):
        for p in plugins.values():
            if p.loaded:
                p.bind_menu_and_methods()
        PluginManager.plugins = plugins

    def start(self):
        logger.debug("Started 'app.application.FlikaApplication.start()'")
        self.show()
        self.raise_()
        QtWidgets.qApp.processEvents()
        logger.debug("Started 'app.application.FlikaApplication.send_user_stats()'")
        self.send_user_stats_thread = Send_User_Stats_Thread()
        self.send_user_stats_thread.start()
        logger.debug("Completed 'app.application.FlikaApplication.send_user_stats()'")
        logger.debug("Completed 'app.application.FlikaApplication.start()'")
        #if 'PYCHARM_HOSTED' not in os.environ and 'SPYDER_SHELL_ID' not in os.environ:
        #    return self.app.exec_()

    def setWindowSize(self):
        #desktop = QtWidgets.QApplication.desktop()
        #width_px=int(desktop.logicalDpiX()*3.4)
        #height_px=int(desktop.logicalDpiY()*.9)
        #self.setGeometry(QtCore.QRect(15, 33, width_px, height_px))
        #self.setFixedSize(326, 80)
        #self.setMaximumSize(width_px*3, 120)
        self.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum);
        self.move(0, 0)

    def _make_menu(self):
        logger.debug("Started 'app.application.FlikaApplication._make_menu()'")
        from ..roi import open_rois
        from ..process.file_ import open_file, open_file_from_gui, open_image_sequence_from_gui, open_points, save_file, save_movie_gui, save_points, save_rois
        fileMenu = self.menuBar().addMenu('File')
        openMenu = fileMenu.addMenu("Open")
        openMenu.addAction("Open Image/Movie", open_file_from_gui)
        openMenu.addAction("Open Image Sequence", open_image_sequence_from_gui)
        openMenu.addAction("Open ROIs", open_rois)
        openMenu.addAction("Open Points", open_points)
        self.recentFileMenu = fileMenu.addMenu('Recent Files')
        self.recentFileMenu.aboutToShow.connect(self._make_recents)
        self.recentFileMenu.triggered.connect(lambda a: open_file(a.text()))
        saveMenu = fileMenu.addMenu("Save")
        saveMenu.addAction("Save Image", save_file)
        saveMenu.addAction("Save Movie (.mp4)", save_movie_gui)
        saveMenu.addAction("Save Points", save_points)
        saveMenu.addAction("Save All ROIs", save_rois)

        fileMenu.addAction("Settings", SettingsEditor.show)
        fileMenu.addAction("&Quit", self.close)#app.quit)


        for menu in g.menus:
            self.menuBar().addMenu(menu)

        self.pluginMenu = self.menuBar().addMenu('Plugins')
        self.pluginMenu.aboutToShow.connect(self._make_plugin_menu)

        self.scriptMenu = self.menuBar().addMenu('Scripts')
        self.scriptMenu.aboutToShow.connect(self._make_script_menu)

        helpMenu = self.menuBar().addMenu("Help")
        url = 'http://flika-org.github.io'
        helpMenu.addAction("Documentation", lambda: QtGui.QDesktopServices.openUrl(QtCore.QUrl(url)))
        helpMenu.addAction("Check For Updates", checkUpdates)
        logger.debug("Completed 'app.application.FlikaApplication._make_menu()'")

    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__[item]
        elif item in g.__dict__:
            return g.__dict__[item]
        raise AttributeError(item)


    def _make_tools(self):
        self.freehand.clicked.connect(lambda: g.settings.__setitem__('mousemode', 'freehand'))
        self.line.clicked.connect(lambda: g.settings.__setitem__('mousemode', 'line'))
        self.rect_line.clicked.connect(lambda: g.settings.__setitem__('mousemode', 'rect_line'))
        self.pencil.clicked.connect(lambda: g.settings.__setitem__('mousemode', 'pencil'))
        self.rectangle.clicked.connect(lambda: g.settings.__setitem__('mousemode', 'rectangle'))
        self.point.clicked.connect(lambda: g.settings.__setitem__('mousemode', 'point'))
        self.mouse.clicked.connect(lambda: g.settings.__setitem__('mousemode', 'mouse'))

        self.point.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.point.customContextMenuRequested.connect(pointSettings)
        self.pencil.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.pencil.customContextMenuRequested.connect(pencilSettings)
        self.rectangle.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.rectangle.customContextMenuRequested.connect(rectSettings)

    def _make_script_menu(self):
        logger.debug('Making script editor')
        self.scriptMenu.clear()
        self.scriptEditorAction = self.scriptMenu.addAction('Script Editor', ScriptEditor.show)
        self.scriptMenu.addSeparator()
        def openScript(script):
            return lambda : ScriptEditor.importScript(script)
        for recent_script in g.settings['recent_scripts']:
            self.scriptMenu.addAction(recent_script, openScript(recent_script))
        logger.debug('Script editor complete')

    def _make_plugin_menu(self):
        logger.debug('Making Plugin Manager')
        self.pluginMenu.clear()
        self.pluginMenu.addAction('Plugin Manager', PluginManager.show)
        self.pluginMenu.addSeparator()
        logger.debug('Plugin Manager complete')

        installedPlugins = [plugin for plugin in PluginManager.plugins.values() if plugin.installed]
        for plugin in sorted(installedPlugins, key=lambda a: -a.lastModified()):
            if isinstance(plugin.menu, QtWidgets.QMenu):
                self.pluginMenu.addMenu(plugin.menu)

    def _make_recents(self):
        logger.debug('Making recent files')
        self.recentFileMenu.clear()
        g.settings['recent_files'] = [f for f in g.settings['recent_files'] if os.path.exists(f)]
        if len(g.settings['recent_files']) == 0:
            noAction = QtWidgets.QAction('No Recent Files', self.recentFileMenu)
            noAction.setEnabled(False)
            self.recentFileMenu.addAction(noAction)
        else:
            for name in g.settings['recent_files'][::-1]:
                if isinstance(name, str) and os.path.exists(name):
                    self.recentFileMenu.addAction(name)
        logger.debug('Recent files complete')

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()   # must accept the dragEnterEvent or else the dropEvent can't occur !!!
        else:
            event.ignore()

    def dropEvent(self, event):
        from ..process.file_ import open_file
        if event.mimeData().hasUrls():   # if file or link is dropped
            for url in event.mimeData().urls():
                filename = url.toLocalFile()
                filename = str(filename)
                #if platform.system() == 'Windows':
                #    filename = filename.split('file:///')[1]
                #else:
                #    filename = filename.split('file://')[1]
                print("filename = '{}'".format(filename))
                open_file(filename)  # This fails on windows symbolic links.  http://stackoverflow.com/questions/15258506/os-path-islink-on-windows-with-python
                event.accept()
        else:
            event.ignore()

    def clear(self):
        """clear(self)
        Close all dialogs, trace windows, and windows
        """
        while g.dialogs:
            g.dialogs.pop(0).close()
        while g.traceWindows:
            g.traceWindows.pop(0).close()
        while g.windows:
            g.windows.pop(0).close()

    def closeEvent(self, event):
        """closeEvent(self, event)
        Close all widgets and exit flika
        """
        print('Closing flika')
        event.accept()
        ScriptEditor.close()
        PluginManager.close()
        self.clear()
        g.settings.save()
        if g.m == self:
            g.m = None

class SetCurrentWindowSignal(QtWidgets.QWidget):
    sig=QtCore.Signal()

    def __init__(self,parent):
        super(SetCurrentWindowSignal, self).__init__(parent)
        self.hide()
logger.debug("Completed 'reading app/application.py'")


"""

My preferred startup sequence for running flika inside an IDE is this:

from flika import *
start_flika()
from flika.utils.imports import *
"""

from ..window import Window
from ..process import *
from ..roi import open_rois
import numpy as np
from pyqtgraph import plot, show
import pyqtgraph as pg
import pyqtgraph.opengl as gl


# -*- coding: utf-8 -*-
"""This module declares Region Of Interest (ROI) types as extensions of pyqtgraph.ROI objects.

Current ROI types are:
    * line
    * rectangle
    * freehand
    * rect_line

Example:
    >>> roi = makeROI('line', [[10, 10], [20, 15]])
    >>> roi.plot()
    >>> roi.copy()
    >>> win2 = open_file()
    >>> roi2 = win2.paste()

Todo:
    * Correct ROI line handles to be at center of pixel
"""
from .logger import logger
logger.debug("Started 'reading roi.py'")
import os
from qtpy import QtGui, QtCore, QtWidgets
import pyqtgraph as pg
import numpy as np
from pyqtgraph.graphicsItems.ROI import Handle
from . import global_vars as g
from .utils.misc import random_color, open_file_gui, nonpartial
class ROI_Drawing(pg.GraphicsObject):
    """Graphics Object for ROIs while initially being drawn. Extends pyqtrgaph.GraphicsObject

    Only used on initial mouse drag, the drawFinished method returns the
    resulting ROI object which can then be modified

    Attributes:
        window (window.Window): Window object to draw the ROI in
        x (int): x coordinate of mouse press
        y (int): y coordinate of mouse press
        kind (str): one of ['rectangle', 'line', 'freehand', 'rect_line']
        color (QtGui.QColor): pen color to draw ROI with
    """
    def __init__(self, window, x, y, kind):
        pg.GraphicsObject.__init__(self)
        window.imageview.addItem(self)
        self.window = window
        self.pts = [pg.Point(round(x), round(y))]
        self.kind = kind
        self.state = {'pos': pg.Point(x, y), 'size': pg.Point(0, 0)}
        self.color = QtGui.QColor(g.settings['roi_color']) if g.settings['roi_color'] != 'random' else random_color()

    def cancel(self):
        g.win.imageview.removeItem(self)
        g.win.currentROI = None
        self.deleteLater()

    def extend(self, x, y):
        new_pt = pg.Point(round(x), round(y))
        if self.kind == 'freehand':
            if self.pts[-1] != new_pt:
                self.pts.append(new_pt)
        elif self.kind in ('line', 'rectangle', 'rect_line'):
            if len(self.pts) == 1:
                self.pts.append(new_pt)
            else:
                self.pts[1] = new_pt

            #self.pts = sorted(self.pts, key=lambda a: a.x()/a.y())

        self.state['pos'] = pg.Point(*np.min(self.pts, 0))
        self.state['size'] = pg.Point(*np.ptp(self.pts, 0))

        self.prepareGeometryChange()
        self.update()

    def paint(self, p, *args):
        pen = QtGui.QPen(self.color)
        pen.setWidth(0)
        p.setPen(pen)
        if self.kind == 'freehand':
            p.drawPolyline(*self.pts)
        elif self.kind == 'rectangle':
            p.drawRect(self.boundingRect())
        elif self.kind in ('rect_line', 'line'):
            p.drawLine(*self.pts)

    def drawFinished(self):
        self.window.imageview.removeItem(self)
        if self.kind == 'rectangle':
            pts = [self.state['pos'], self.state['size']]
        else:
            pts = self.pts
        return makeROI(self.kind, pts, self.window, color=self.color)


    def contains(self, *args):
        if len(args) == 2:
            args = [pg.Point(*args)]
        return pg.GraphicsObject.contains(self, *args)

    def boundingRect(self):
        return QtCore.QRectF(self.state['pos'].x(), self.state['pos'].y(), self.state['size'].x(), self.state['size'].y())

class ROI_Base():
    """ROI_Base interface for all ROI types

    Template class for common and abstract functions, connects window.closeEvent to pyqtgraph.ROI.delete, set the window.currentROI to self

    Attributes:
        colorDialog: dialog for selecting the color of the ROI and its trace
        traceWindow: the :class:`TraceFig <flika.tracefig.TraceFig>` that this ROI is plotted to, or None
        pts: array of XY values used to copy the ROI
        linkedROIs: set of rois that act as one ROI

    Note:
        All ROI objects implement the following methods:
            getMask():
                returns the [2, N] array of mask coordinates
            getPoints():
                returns the [N, 2] points that make up the ROI   
            draw_from_points(pts, finish=True):
                updates the point locations that make the ROI, used in linked ROIs

    """
    INITIAL_ARGS = {'translateSnap': True, 'removable': True, 'snapSize': 1, 'scaleSnap': True}

    def __init__(self, window, pts):
        self.window = window #: window.Window: Parent window that this ROI belongs to
        self.colorDialog=QtWidgets.QColorDialog()
        self.colorDialog.colorSelected.connect(self.colorSelected)
        self.window.closeSignal.connect(self.delete)
        self.window.currentROI = self
        self.traceWindow = None  #: tracefig.TraceFig: the Trace window that this ROI is plotted in. To test if roi is plotted, check 'roi.traceWindow is None'
        self.pts = np.array(pts) #: list: Array of points that make up the boundary of the ROI
        self.linkedROIs = set()
        self.resetSignals()
        self.makeMenu()

    def mouseClickEvent(self, ev):
        self.window.currentROI = self
        super().mouseClickEvent(ev)

    def resetSignals(self):
        try:
            self.sigRegionChanged.disconnect()
        except:
            pass
        try:
            self.sigRegionChangeFinished.disconnect()
        except:
            pass
        self.sigRegionChanged.connect(self.onRegionChange)
        self.sigRegionChangeFinished.connect(self.onRegionChangeFinished)

    def updateLinkedROIs(self, finish=False):
        for roi in self.linkedROIs:
            roi.blockSignals(True)
            roi.draw_from_points(self.pts, finish=False)
            if roi.traceWindow is not None:
                if not finish:
                    roi.traceWindow.translated(roi)
                else:
                    roi.traceWindow.translateFinished(roi)
            roi.blockSignals(False)

    def redraw_trace(self):
        """Emit the translateFinished signal which redraws the ROI trace
        """
        if self.traceWindow is not None:
            self.traceWindow.translateFinished(self)

    def onRegionChange(self):
        self.pts = self.getPoints()
        self.updateLinkedROIs(finish=False)
        
    def onRegionChangeFinished(self):
        self.pts = self.getPoints()
        self.updateLinkedROIs(finish=True)

    def link(self,roi):
        '''Link this roi to another, so a translation of one will cause a translation of the other'''
        if not isinstance(roi, type(self)):
            return
        join = self.linkedROIs | roi.linkedROIs | {self, roi}
        self.linkedROIs = join - {self}
        roi.linkedROIs = join - {roi}

    def getMask(self):
        '''Returns the list of integer points contained within the ROI, differs by ROI type
        '''
        raise NotImplementedError()

    def getTrace(self, bounds=None):
        '''Compute the average of the pixels within this ROI in its window

        Returns:
            Average value within ROI mask, as an array. Cropped to bounds if specified
        '''
        trace = None
        if self.window.image.ndim == 4 or self.window.metadata['is_rgb']:
            g.alert("Plotting trace of RGB movies is not supported. Try splitting the channels.")
            return None
        s1, s2 = self.getMask()
        if np.size(s1) == 0 or np.size(s2) == 0:
            trace = np.zeros(self.window.mt)

        elif self.window.image.ndim == 3:
            trace = self.window.image[:, s1, s2]
            while trace.ndim > 1:
                trace = np.average(trace, 1)
        elif self.window.image.ndim == 2:
            trace = self.window.image[s1, s2]
            trace = [np.average(trace)]

        if bounds:
            trace = trace[bounds[0]:bounds[1]]
        return trace

    def getPoints(self):
        '''Get points that represent this ROI, used for exporting
        '''
        raise NotImplementedError()

    def draw_from_points(self, pts, finish=True):
        '''Redraw the ROI from the given points, used on linked ROIs

        Args:
            pts: points used to represent ROI, often handle positions
            finish: whether or not to emit the onRegionChangeFinished signal
        '''
        raise NotImplementedError()

    def setMouseHover(self, hover):
        """
        Inform the ROI that the mouse is or is not hovering over it.

        Args:
            hover (bool)

        """
        if self.mouseHovering is hover:
            return
        self.mouseHovering = hover
        if hover:
            self.currentPen = pg.mkPen(QtGui.QColor(255, 0, 0))
        else:
            self.currentPen = self.pen

        self.update()

    def plot(self):
        """Plot the ROI trace in a :class:`TraceFig <flika.tracefig.TraceFig>`

        Returns:
            tracefig.TraceFig: the trace window that the ROI was plotted to
        """
        from .tracefig import roiPlot
        self.traceWindow = roiPlot(self)
        if self.traceWindow == None:
            return
        self.traceWindow.indexChanged.connect(self.window.setIndex)
        self.plotSignal.emit()
        return self.traceWindow

    def changeColor(self):
        self.colorDialog.open()
        
    def colorSelected(self, color):
        """Set the pen color of the ROI

        Args:
            color (QtGui.QColor): new color for the ROI
        """
        if color.isValid():
            self.setPen(QtGui.QColor(color.name()))
            self.sigRegionChangeFinished.emit(self)

    def unplot(self):
        """Remove the ROI from its :class:`TraceFig <flika.tracefig.TraceFig>`
        """
        try:
            self.traceWindow.indexChanged.disconnect(self.window.setIndex)
        except:
            # sometimes errors, says signals not connected
            pass
        if self.traceWindow != None:
            self.traceWindow.removeROI(self)
        
        self.traceWindow = None

    def copy(self):
        """Store this ROI in the clipboard
        """
        g.clipboard=self

    def raiseContextMenu(self, ev):
        pos = ev.screenPos()
        self.menu.addSeparator()
        self.menu.addActions(self.window.menu.actions())
        self.menu.popup(QtCore.QPoint(pos.x(), pos.y()))
    
    def makeMenu(self):
        def plotPressed():
            if plotAct.text() == "&Plot":
                self.plot()
            else:
                self.unplot()

        plotAct = QtWidgets.QAction("&Plot", self, triggered=plotPressed)
        colorAct = QtWidgets.QAction("&Change Color",self,triggered=self.changeColor)
        copyAct = QtWidgets.QAction("&Copy", self, triggered=self.copy)
        remAct = QtWidgets.QAction("&Delete", self, triggered=self.delete)
        self.menu = QtWidgets.QMenu("ROI Menu")

        def updateMenu():
            #plotAct.setEnabled(self.window.image.ndim > 2)
            plotAct.setText("&Plot" if self.traceWindow == None else "&Unplot")
            self.window.menu.aboutToShow.emit()

        self.menu.addAction(plotAct)
        self.menu.addAction(colorAct)
        self.menu.addAction(copyAct)
        self.menu.addAction(remAct)
        self.menu.aboutToShow.connect(updateMenu)

    def delete(self):
        """Remove the ROI from its window, unlink all ROIs and delete the object"""
        self.unplot()
        for roi in self.linkedROIs:
            if self in roi.linkedROIs:
                roi.linkedROIs.remove(self)
        if self in self.window.rois:
            self.window.rois.remove(self)
        self.window.currentROI=None
        self.window.imageview.removeItem(self)
        self.window.closeSignal.disconnect(self.delete)
        if g.clipboard == self:
            g.clipboard = None

    def drawFinished(self):
        self.window.imageview.addItem(self)
        self.window.rois.append(self)
        self.window.currentROI = self

    def _str(self):
        """Return ROI kind and points for easy export and import
    
        Returns:
            str: ROI string representation
        """
        s = self.kind + '\n'
        for x, y in self.pts:
            s += '{} {}\n'.format(x, y)
        return s

    def showMask(self):
        """Create a new binary window that visualizes the ROI mask

        Returns:
            window.Window: created mask window
        """
        from .window import Window
        self.copy()
        im = np.zeros_like(self.window.imageview.getImageItem().image)
        s1, s2 = self.getMask()
        im[s1, s2] = 1
        w = Window(im)
        w.paste()
        return w


class ROI_line(ROI_Base, pg.LineSegmentROI):
    '''ROI Line class for selecting a straight line of pixels between two points.

    Extends from :class:`ROI_Base <flika.roi.ROI_Base>` and pyqtgraph pyqtgraph.LineSegmentROI

    Attributes:
        kymograph (Window): :class:`Window <flika.window.Window>` showing 2d kymograph.
    '''
    kind = 'line'
    plotSignal = QtCore.Signal()
    
    def __init__(self, window, positions, **kargs):
        roiArgs = self.INITIAL_ARGS.copy()
        roiArgs.update(kargs)
        pg.LineSegmentROI.__init__(self, positions=positions, **roiArgs)
        self.kymograph = None
        self.kymographAct = QtWidgets.QAction("&Kymograph", self, triggered=self.update_kymograph)
        ROI_Base.__init__(self, window, positions)
        #self.snapPoints()

    def paint(self, p, *args):
        p.setRenderHint(QtGui.QPainter.Antialiasing)
        p.setPen(self.currentPen)
        h1 = self.handles[0]['item'].pos()
        h2 = self.handles[1]['item'].pos()
        p.drawLine(h1, h2)

    def resetSignals(self):
        ROI_Base.resetSignals(self)
        self.sigRegionChanged.connect(self.snapPoints)

    def snapPoints(self):
        """Correct ROI points to be at the center of pixels, for clarity
        """
        fix = False
        self.blockSignals(True)
        for handle in self.handles:
            pos = handle['pos']
            pos_snap = self.getSnapPosition(pg.Point(pos))# + pg.Point(.5, .5)
            if not (pos == pos_snap):
                handle['item'].setPos(pos_snap)
                handle['pos'] = pos_snap
                fix = True

        self.blockSignals(False)
        #if fix:
        #    self.sigRegionChanged.emit(self)

    def draw_from_points(self, pts, finish=True):
        self.blockSignals(True)
        self.movePoint(self.handles[0]['item'], pts[0], finish=False)
        self.movePoint(self.handles[1]['item'], pts[1], finish=False)
        self.pts = pts
        self.blockSignals(False)
        if finish:
            self.sigRegionChangeFinished.emit(self)

    def delete(self):
        ROI_Base.delete(self)
        if self.kymograph:
            self.deleteKymograph()

    def getMask(self):
        from skimage.draw import line
        x=np.array([p[0] for p in self.pts], dtype=int)
        y=np.array([p[1] for p in self.pts], dtype=int)
        xx, yy = line(x[0],y[0],x[1],y[1])
        idx_to_keep = np.logical_not( (xx>=self.window.mx) | (xx<0) | (yy>=self.window.my) | (yy<0))
        xx = xx[idx_to_keep]
        yy = yy[idx_to_keep]
        return xx, yy

    def getPoints(self):
        return np.array([handle['pos'] + self.state['pos'] for handle in self.handles])

    def makeMenu(self):
        ROI_Base.makeMenu(self)
        self.menu.addAction(self.kymographAct)
        self.kymographAct.setEnabled(self.window.image.ndim == 3 and not self.window.metadata['is_rgb'])

    def update_kymograph(self):
        tif=self.window.image
        if tif.ndim != 3:
            g.alert("Can only kymograph a 3d movie")
            return

        xx, yy = self.getMask()
        mt = len(tif)
        if len(xx) == 0:
            return
        xx = np.array(xx)
        yy = np.array(yy)
        
        if len(xx) == 0:
            return
        mn=np.zeros((mt,len(xx)))
        for t in np.arange(mt):
            mn[t]=tif[t,xx,yy]
        mn=mn.T
        if self.kymograph is None:
            self.createKymograph(mn)
        else:
            self.kymograph.imageview.setImage(mn,autoLevels=False,autoRange=False)
            #self.kymograph.imageview.view.setAspectLocked(lock=True,ratio=mn.shape[1]/mn.shape[0])
    
    def createKymograph(self,mn):
        from .window import Window
        oldwindow = g.win
        name=oldwindow.name+' - Kymograph'
        self.kymograph=Window(mn,name,metadata=self.window.metadata)
        self.sigRegionChanged.connect(self.update_kymograph)
        self.kymograph.closeSignal.connect(self.deleteKymograph)
        self.sigRemoveRequested.connect(self.deleteKymograph)

    def deleteKymograph(self):
        self.kymograph.closeSignal.disconnect(self.deleteKymograph)
        self.sigRegionChanged.disconnect(self.update_kymograph)
        self.kymograph=None

class ROI_rectangle(ROI_Base, pg.ROI):
    '''ROI rectangle class for selecting a set width and height group of pixels on an image.

    Extends from :class:`ROI_Base <flika.roi.ROI_Base>` and pyqtgraph.ROI
    '''
    kind = 'rectangle'
    plotSignal = QtCore.Signal()

    def __init__(self, window, pos, size, resizable=True, **kargs):
        """__init__ of ROI_rectangle class

        Args:
            pos (2-tuple): position of top left corner
            size: (2-tuple): width and height of the rectangle
            resizable (bool): add resize handles to ROI, this cannot be changed after creation
        """
        roiArgs = self.INITIAL_ARGS.copy()
        roiArgs.update(kargs)
        pos = np.array(pos, dtype=int)
        size = np.array(size, dtype=int)

        pg.ROI.__init__(self, pos, size, **roiArgs)
        if resizable:
            self.addScaleHandle([0, 1], [1, 0])
            self.addScaleHandle([1, 0], [0, 1])
            self.addScaleHandle([0, 0], [1, 1])
            self.addScaleHandle([1, 1], [0, 0])
        self.cropAction = QtWidgets.QAction('&Crop', self, triggered=self.crop)
        ROI_Base.__init__(self, window, [pos, size])

    def center_around(self, x, y):
        """Relocate ROI so center lies at Point (x, y). size is not changed

        Args:
            x (int): new center for rectangle on X axis
            y (int): new center for rectangle on Y axis
        """
        old_pts = self.getPoints()
        old_center = old_pts[0] + .5 * old_pts[1]
        new_center = np.array([x, y])
        diff = new_center - old_center
        new_pts = np.array([old_pts[0]+diff, old_pts[1]])
        self.draw_from_points(new_pts)

    def getPoints(self):
        return np.array([self.state['pos'], self.state['size']], dtype=int)

    def contains_pts(self, x, y):
        target = np.array([x, y])
        return np.all(self.pts[0] < target) and np.all(target < self.pts[0]+self.pts[1])

    def getMask(self):
        x, y = self.state['pos']
        ww, hh = self.state['size']

        xmin = max(x, 0)
        ymin = max(y, 0)
        xmax = min(x+ww, self.window.mx)
        ymax = min(y+hh, self.window.my)

        xx, yy = np.meshgrid(np.arange(xmin, xmax, dtype=int), np.arange(ymin, ymax, dtype=int))

        return xx.flatten(), yy.flatten()

    def draw_from_points(self, pts, finish=True):
        self.setPos(pts[0], finish=False)
        self.setSize(pts[1], finish=False)
        self.pts = np.array(pts)
        self.sigRegionChanged.emit(self)
        if finish:
            self.sigRegionChangeFinished.emit(self)

    def makeMenu(self):
        ROI_Base.makeMenu(self)
        self.menu.addAction(self.cropAction)

    def crop(self):
        """Create a new window of the image cropped to this ROI

        Returns:
            window.Window: cropped image Window
        """
        from .window import Window
        r = self.boundingRect()
        p1 = r.topLeft() + self.state['pos']
        p2 = r.bottomRight() + self.state['pos']
        x1, y1 = int(p1.x()), int(p1.y())
        x2, y2 = int(p2.x()), int(p2.y())

        if x1<0: x1=0
        if y1<0: y1=0

        tif=self.window.image
        #if self.window.imageview.hasTimeAxis():
        if self.window.mt > 1:
            mt, mx, my = tif.shape[:3]
            if x2>=mx: x2=mx-1
            if y2>=my: y2=my-1
            newtif=tif[:,x1:x2,y1:y2]
        else:
            mx, my = tif.shape[:2]
            if x2>=mx: x2=mx-1
            if y2>=my: y2=my-1
            newtif=tif[x1:x2,y1:y2]

        w =  Window(newtif, self.window.name+' Cropped', metadata=self.window.metadata)
        w.imageview.setImage(newtif, axes=self.window.imageview.axes)
        w.image = newtif
        return w

class ROI_freehand(ROI_Base, pg.ROI):
    """ROI freehand class for selecting a polygon from the original image.

    Extends from :class:`ROI_Base <flika.roi.ROI_Base>` and pyqtgraph.ROI.
    """
    kind = 'freehand'
    plotSignal = QtCore.Signal()
    def __init__(self, window, pts, **kargs):
        roiArgs = self.INITIAL_ARGS.copy()
        roiArgs.update(kargs)
        roiArgs['closed'] = True
        pg.ROI.__init__(self, np.min(pts, 0), np.ptp(np.array(pts), 0), translateSnap=(1, 1), **kargs)
        ROI_Base.__init__(self, window, pts)
        self._untranslated_pts = np.subtract(self.pts, self.pos())
        self._untranslated_mask = None
        self.getMask()

    def shape(self):
        p = QtGui.QPainterPath()
        p.moveTo(*self._untranslated_pts[0])
        for i in range(len(self._untranslated_pts)):
            p.lineTo(*self._untranslated_pts[i])
        p.lineTo(*self._untranslated_pts[0])
        return p

    def paint(self, painter, *args):
        painter.setPen(self.currentPen)
        painter.drawPolygon(*[pg.Point(a, b) for a, b in self._untranslated_pts])

    def draw_from_points(self, pts, finish=False):
        self.blockSignals(True)
        self.setPos(*np.min(pts, 0), False)
        self.setSize(np.ptp(pts, 0), False)
        self._untranslated_pts = np.subtract(pts, self.pos())
        self.pts = pts
        
        self.sigRegionChanged.emit(self)
        if finish:
            self.sigRegionChangeFinished.emit(self)
        self.blockSignals(False)

    def contextMenuEnabled(self):
        return True

    def getPoints(self):
        x, y = self.state['pos']
        return np.add(self._untranslated_pts, [x, y])

    def contains_pt(self, x, y):
        pt = QtCore.QPointF(x, y) - self.pos()
        qPainterPath = self.shape()
        return qPainterPath.contains(pt)

    def contains_pts(self, x, y):
        ''' not yet implemented'''
        pass

    def getMask(self):
        from skimage.draw import polygon
        if self._untranslated_mask is not None:
            xx = self._untranslated_mask[0] + int(self.state['pos'][0])
            yy = self._untranslated_mask[1] + int(self.state['pos'][1])
        else:
            x, y = np.transpose(self._untranslated_pts)
            mask=np.zeros(self.window.imageDimensions())
            xx, yy = polygon(x,y,shape=mask.shape)
            self._untranslated_mask = xx, yy

        idx_to_keep = np.logical_not( (xx>=self.window.mx) | (xx<0) | (yy>=self.window.my) | (yy<0))
        xx = xx[idx_to_keep]
        yy = yy[idx_to_keep]
        return xx, yy

class ROI_rect_line(ROI_Base, QtWidgets.QGraphicsObject):
    """Collection of linked line segments with adjustable width.

    Extends from :class:`ROI_Base <flika.roi.ROI_Base>` and QtWidgets.QGraphicsObject

    Attributes:
        kymograph (Window): :class:`Window <flika.window.Window>` showing 2d kymograph.
    """
    kind = 'rect_line'
    plotSignal = QtCore.Signal()
    sigRegionChanged = QtCore.Signal(object)
    sigRegionChangeFinished = QtCore.Signal(object)

    def __init__(self, window, pts, width=1, **kargs):
        self.roiArgs = self.INITIAL_ARGS.copy()
        self.roiArgs.update(kargs)
        self.roiArgs['scaleSnap'] = False
        self.width = width
        self.kymograph = None
        QtWidgets.QGraphicsObject.__init__(self)
        self.kymographAct = QtWidgets.QAction("&Kymograph", self, triggered=self.update_kymograph)
        self.removeLinkAction = QtWidgets.QAction('Remove Last Link', self, triggered=nonpartial(self.removeSegment))
        self.setWidthAction = QtWidgets.QAction("Set Width", self, triggered=nonpartial(self.setWidth))
        ROI_Base.__init__(self, window, pts)
        self.getPoints = self.getHandlePositions
        self.pen = QtGui.QPen(QtGui.QColor(255, 255, 0))
        self.pen.setWidth(0)
        self.lines = []
        if len(pts) < 2:
            raise Exception("Must start with at least 2 points")
        self.extendHandle = None
            
        self.addSegment(pts[1], connectTo=pts[0])
        for p in pts[2:]:
            self.addSegment(p)

    def getHandles(self):
        handles = []
        for line in self.lines:
            handles.append(line.getHandles()[0])
        handles.append(self.lines[-1].getHandles()[1])
        return handles

    def movePoint(self, handle, *args, **kargs):
        if isinstance(handle, int):
            handle = self.getHandles()[handle]
        kargs['finish'] = False
        self.blockSignals(True)
        for line in handle.rois:
            line.movePoint(handle, *args, **kargs)
        self.blockSignals(False)
        self.sigRegionChanged.emit(self)
        self.sigRegionChangeFinished.emit(self)

    def delete(self):
        ROI_Base.delete(self)
        if self.kymograph:
            self.deleteKymograph()

    def draw_from_points(self, pts, finish=False):
        while len(self.lines) > 1:
            self.removeSegment(self.lines[-1])

        self.lines[0].movePoint(0, pts[0])
        self.lines[0].movePoint(1, pts[1])
        for p in pts[2:]:
            self.addSegment(p)

        if finish:
            self.sigRegionChangeFinished.emit(self)
        self.pts = pts

    def getTrace(self, bounds=None):
        if self.window.image.ndim > 3 or self.window.metadata['is_rgb']:
            g.alert("Plotting trace of RGB movies is not supported. Try splitting the channels.")
            return None
        if self.window.image.ndim == 3:
            region = self.getArrayRegion(self.window.imageview.image, self.window.imageview.getImageItem(), (1, 2))
            while region.ndim > 1:
                region = np.average(region, 1)
        elif self.window.image.ndim == 2:
            region = self.getArrayRegion(self.window.imageview.image, self.window.imageview.getImageItem(), (0, 1))
            region = [np.average(region)]

        if bounds:
            region = region[bounds[0]:bounds[1]]
        return region

    def preview(self):
        im = self.getArrayRegion(self.window.imageview.getImageItem().image, self.window.imageview.getImageItem(), (0, 1))
        if not hasattr(self, 'prev'):
            from .window import Window
            self.prev = Window(im)
            self.sigRegionChanged.connect(lambda a: self.preview())
        else:
            self.prev.imageview.setImage(im)

    def getHandlePositions(self):
        """Return the positions of all handles in local coordinates."""
        p = self.mapFromScene(self.lines[0].getHandles()[0].scenePos())
        p = self.lines[0].getSnapPosition([p.x(), p.y()])
        pos = [p]
        for l in self.lines:
            p = self.mapFromScene(l.getHandles()[1].scenePos())
            p = l.getSnapPosition([p.x(), p.y()])
            pos.append(p)
        self.pts = pos
        return self.pts

    def getArrayRegion(self, arr, img=None, axes=(0,1), **kwds):
        rgns = []
        for l in self.lines:
            rgn = l.getArrayRegion(arr, img, axes=axes, **kwds)
            if rgn is None:
                continue
                #return None
            rgns.append(rgn)
            #print l.state['size']
            
        ## make sure orthogonal axis is the same size
        ## (sometimes fp errors cause differences)
        if img.axisOrder == 'row-major':
            axes = axes[::-1]
        ms = min([r.shape[axes[1]] for r in rgns])
        sl = [slice(None)] * rgns[0].ndim
        sl[axes[1]] = slice(0,ms)
        rgns = [r[sl] for r in rgns]
        #print [r.shape for r in rgns], axes
        
        return np.concatenate(rgns, axis=axes[0])
        
    def addSegment(self, pos=(0, 0), connectTo=None):
        """
        Add a new segment to the ROI connecting from the previous endpoint to *pos*.
        (pos is specified in the parent coordinate system of the MultiRectROI)
        """
        ## by default, connect to the previous endpoint
        if len(self.lines) == 0 or connectTo == self.lines[0].getHandles()[0]:
            ind = 0
        else:
            ind = len(self.lines)

        def dragEvent(handle, ev):
            pos = self.window.imageview.view.mapToView(ev.scenePos())
            if ev.button() == QtCore.Qt.RightButton:
                edgeHandles = self.getHandles()
                edgeHandles = edgeHandles[0], edgeHandles[-1]
                if self.extendHandle is None and handle not in edgeHandles:
                    return

                if len(handle.rois) == 1 and ev.isStart():
                    _, self.extendHandle = self.addSegment(pos, connectTo=handle)
                elif ev.isFinish():
                    self.extendFinished()
                else:
                    self.extend(round(pos.x()), round(pos.y()), finish=False)
            else:
                for roi in handle.rois:
                    roi.movePoint(handle, [round(pos.x()), round(pos.y())])
                return
            Handle.mouseDragEvent(handle, ev)
            ev.accept()

        if connectTo is None:
            connectTo = self.lines[-1].getHandles()[1]

        ## create new ROI
        if len(self.lines) > 0:
            newRoi = pg.ROI((0,0), [1, self.width], parent=self, pen=self.pen, **self.roiArgs)
        else:
            newRoi = pg.ROI((0,0), [1, self.width], pen=self.pen, **self.roiArgs)
            newRoi.setParentItem(self)

        if len(self.lines) == 0 or ind > 0: # add handles in order
            ## Add first SR handle
            if isinstance(connectTo, Handle):
                h = newRoi.addScaleRotateHandle([0, 0.5], [1, 0.5], item=connectTo)
                newRoi.movePoint(connectTo, connectTo.scenePos(), coords='scene')
            else:
                h = newRoi.addScaleRotateHandle([0, 0.5], [1, 0.5])
                newRoi.movePoint(h, connectTo, coords='scene')
            h.mouseDragEvent = lambda ev: dragEvent(h, ev)

            ## add second SR handle
            h2 = newRoi.addScaleRotateHandle([1, 0.5], [0, 0.5])
            h2.mouseDragEvent = lambda ev: dragEvent(h2, ev)
            newRoi.movePoint(h2, pos)
        else:
            h2 = newRoi.addScaleRotateHandle([1, 0.5], [0, 0.5])
            h2.mouseDragEvent = lambda ev: dragEvent(h2, ev)
            newRoi.movePoint(h2, pos)

            if isinstance(connectTo, Handle):
                h = newRoi.addScaleRotateHandle([0, 0.5], [1, 0.5], item=connectTo)
                newRoi.movePoint(connectTo, connectTo.scenePos(), coords='scene')
            else:
                h = newRoi.addScaleRotateHandle([0, 0.5], [1, 0.5])
                newRoi.movePoint(h, connectTo, coords='scene')
            h.mouseDragEvent = lambda ev: dragEvent(h, ev)


        self.lines.insert(ind, newRoi)
        self.lines[0]._updateView()
            
        newRoi.translatable = False
        newRoi.hoverEvent = lambda e: self.hoverEvent(newRoi, e)
        newRoi.raiseContextMenu = self.raiseContextMenu
        #newRoi.sigRegionChangeStarted.connect(self.roiChangeStartedEvent) 
        newRoi.sigRegionChanged.connect(lambda a: self.sigRegionChanged.emit(self))
        newRoi.sigRegionChangeFinished.connect( lambda a: self.sigRegionChangeFinished.emit(self))
        self.sigRegionChanged.emit(self)
        return newRoi, h2

    def removeSegment(self, segment=None, finish=True): 
        """Remove a segment from the ROI"""
        if segment is None:
            segment = self.removeLinkAction.data()
            if segment is None:
                return
        elif isinstance(segment, int):
            segment = self.lines[segment]
        
        segment.sigRegionChangeFinished.disconnect()
        segment.sigRegionChanged.disconnect()

        for h in segment.getHandles():
            if len(h.rois) == 2 and h.parentItem() == segment:
                otherROI = [line for line in h.rois if line != segment][0]
                h.setParentItem(otherROI)
                h.setPos(0, .5)
            h.disconnectROI(segment)
        if segment in self.lines:
            self.lines.remove(segment)

        segment.scene().removeItem(segment)
        if len(self.lines) == 0:
            self.delete()
        
    def extend(self, x, y, finish=True):
        self.blockSignals(True)
        point = pg.Point(x, y)

        if self.extendHandle is not None:
            for roi in self.lines:
                if self.extendHandle in roi.getHandles():
                    roi.movePoint(self.extendHandle, point, finish=False)
            #self.extendHandle.movePoint(self.window.imageview.getImageItem().mapToScene(point), finish=False)
        else:
            self.addSegment((x, y))
        self.blockSignals(False)

        self.sigRegionChanged.emit(self)
        if finish:
            self.sigRegionChangeFinished.emit(self)

    def drawFinished(self):
        ROI_Base.drawFinished(self)
        for l in self.lines:
            l._updateView()

    def extendFinished(self):
        self.extendHandle = None
        for l in self.lines: # fix resizing handles. First link Viewbox was set to something different
            l._updateView()
        self.sigRegionChangeFinished.emit(self)

    def hoverEvent(self, l, ev):
        if ev.enter:
            pen = QtGui.QPen(QtGui.QColor(255, 0, 0))
            pen.setWidth(0)
            l.setPen(pen)
            l.mouseHovering = True
            #self.setCurrentPen(pen)
        elif ev.exit:
            l.setPen(self.pen)
            self.setCurrentPen(self.pen)
            l.mouseHovering = False

    def getMask(self):

        xxs = []
        yys = []
        for i in range(len(self.pts)-1):
            p1, p2 = self.pts[i], self.pts[i+1]
            xx, yy = line(int(p1[0]), int(p1[1]), int(p2[0]), int(p2[1]))
            idx_to_keep = np.logical_not( (xx>=self.window.mx) | (xx<0) | (yy>=self.window.my) | (yy<0))
            xx = xx[idx_to_keep]
            yy = yy[idx_to_keep]
            xxs.extend(xx)
            yys.extend(yy)

        return np.array(xxs, dtype=int), np.array(yys, dtype=int)

    def makeMenu(self):
        ROI_Base.makeMenu(self)
        self.menu.addAction(self.removeLinkAction)
        self.menu.addAction(self.setWidthAction)
        self.menu.addAction(self.kymographAct)
        self.kymographAct.setEnabled(self.window.image.ndim > 2)

    def raiseContextMenu(self, ev):
        currentLines = [line for line in self.lines if line.mouseHovering]
        if len(currentLines) > 0 and np.any([len(i.rois)==1 for i in currentLines[0].getHandles()]):
            self.removeLinkAction.setText("Remove Link")
            self.removeLinkAction.setVisible(True)
            self.removeLinkAction.setData(currentLines[0])
        else:
            self.removeLinkAction.setVisible(False)
        
        ROI_Base.raiseContextMenu(self, ev)

    def boundingRect(self):
        return QtCore.QRectF()

    def paint(self, p, *args):
        pass

    def setPen(self, pen):
        pen = QtGui.QPen(pen)
        pen.setWidth(0)
        self.pen = pen
        self.setCurrentPen(pen)

    def setCurrentPen(self, pen):
        pen = QtGui.QPen(pen)
        pen.setWidth(0)
        for l in self.lines:
            l.currentPen = pen
            l.update()

    def update_kymograph(self):
        tif=self.window.image
        if tif.ndim != 3:
            g.alert("Can only kymograph on 3D movies")
            return
        
        if self.width == 1:
            w, h = self.window.imageDimensions()
            r = QtCore.QRect(0, 0, w, h)
            xx, yy = self.getMask()
            mn = tif[:, xx, yy].T
        else:
            region = self.getArrayRegion(self.window.imageview.image, self.window.imageview.getImageItem(), (1, 2))
            mn = np.average(region, 2).T


        if self.kymograph is None:
            self.createKymograph(mn)
        else:
            if mn.size > 0:
                self.kymograph.imageview.setImage(mn,autoLevels=False,autoRange=False)
            #self.kymograph.imageview.view.setAspectLocked(lock=True,ratio=mn.shape[1]/mn.shape[0])

    def setWidth(self, newWidth=None):
        """Set the width of each segment in the ROI

        Args:
            newWidth (int): new width of all segments
        """
        s = True
        if newWidth is None:
            newWidth, s = QtWidgets.QInputDialog.getInt(None, "Enter a width value", 'Float Value', value=self.width)
        if not s or s == 0:
            return
        for l in self.lines:
            l.scale([1.0, newWidth/self.width], center=[0.5,0.5])
        self.width = newWidth
        self.sigRegionChangeFinished.emit(self)

    def createKymograph(self,mn):
        from .window import Window
        oldwindow=g.win
        name=oldwindow.name+' - Kymograph'
        self.kymograph=Window(mn,name,metadata=self.window.metadata)
        self.kymographproxy = pg.SignalProxy(self.sigRegionChanged, rateLimit=1, slot=self.update_kymograph) #This will only update 3 Hz
        self.sigRegionChanged.connect(self.update_kymograph)
        self.kymograph.closeSignal.connect(self.deleteKymograph)

    def deleteKymograph(self):
        self.kymographproxy.disconnect()
        self.sigRegionChanged.disconnect(self.update_kymograph)
        self.kymograph.closeSignal.disconnect(self.deleteKymograph)
        self.kymograph=None

def makeROI(kind, pts, window=None, color=None, **kargs):
    """Create an ROI object in window with the given points

    Args:
        kind (str): one of ['line', 'rectangle', 'freehand', 'rect_line']
        pts ([N, 2] list of coords): points used to draw the ROI, differs by kind
        window (window.Window): window to draw the ROI in, or currentWindow if not specified
        color (QtGui.QColor): pen color of the new ROI
        **kargs: additional arguments to pass to the ROI __init__ function

    Returns:
        ROI Object extending ROI_Base
    """
    if window is None:
        window = g.win
        if window is None:
            g.alert('ERROR: In order to make and ROI a window needs to be selected')
            return None

    if kind == 'freehand':
        roi = ROI_freehand(window, pts, **kargs)
    elif kind == 'rectangle':
        if len(pts) > 2:
            size = np.ptp(pts,0)
            top_left = np.min(pts,0)
        else:
            size = pts[1]
            top_left = pts[0]
        roi = ROI_rectangle(window, top_left, size, **kargs)
    elif kind == 'line':
        roi = ROI_line(window, (pts), **kargs)
    elif kind == 'rect_line':
        roi = ROI_rect_line(window, pts, **kargs)
    else:
        g.alert("ERROR: THIS KIND OF ROI COULD NOT BE FOUND: {}".format(kind))
        return None

    if color is None or not isinstance(color, QtGui.QColor):
        pen = QtGui.QPen(QtGui.QColor(g.settings['roi_color']) if g.settings['roi_color'] != 'random' else random_color())
    else:
        pen = QtGui.QPen(color)
    pen.setWidth(0)

    roi.drawFinished()
    roi.setPen(pen)
    return roi

def open_rois(filename=None):
    """
    Open an roi.txt file, creates ROI objects and places them in the current Window.
    
    Args:
        filename (str): The filename (including full path) of the roi.txt file.

    Returns:
        list of rois

    """
    if filename is None:
        filetypes = '*.txt'
        prompt = 'Load ROIs from file'
        filename = open_file_gui(prompt, filetypes=filetypes)
        if filename is None:
            return None
    text = open(filename, 'r').read()
    rois = []
    kind = None
    pts = None
    for text_line in text.split('\n'):
        if kind is None:
            kind = text_line
            pts = []
        elif text_line == '':
            roi = makeROI(kind,pts)
            rois.append(roi)
            kind = None
            pts = None
        else:
            pts.append(tuple(int(float(i)) for i in text_line.split()))

    return rois
logger.debug("Completed 'reading roi.py'")
